import warnings
import numpy as np
import pandas as pd
import cvxpy as cp
from scipy import stats
from . import utilities
from . import dist_reg
from .utilities import BatchedCategorical
from . import interpolation, generic
from .delta import _delta_bootstrap_ses
from typing import Optional, Union


### Helper functions ###

def compute_cvar(dists, n: int, alpha: float, lower: bool=True, m: int=1000):
	"""
	Computes cvar using quantile approximation with m values.

	Parameters
	----------
	dists : stats.dist
		scipy distribution function of shape n
	n : int
		Batch dimension
	alpha : array or float	
		float or n-length array
	lower : bool
		If true, computes the lower CVAR. 
		Else computes the upper CVAR.
	m : int
		Number of interpolation points
	
	Returns
	-------
	cvars : array
		n-length array.
		E[Y | Y <= Q_{alpha}(Y)] from Y ~ dists if lower = True.
		If lower = False, replaces <= with >=.
	"""

	if isinstance(alpha, float) or isinstance(alpha, int):
		alpha = alpha * np.ones((n))
	# find quantiles
	if lower:
		qs = np.linspace(1/(m+1), alpha, m)
	else:
		qs = np.linspace(alpha, m/(m+1), m)
	# take average
	qmc = dists.ppf(q=qs)
	if qmc.shape != (m, n):
		raise ValueError(f"Unexpected shape of qmc={qmc.shape}")
	cvar_est = qmc.mean(axis=0)
	return cvar_est

def compute_analytical_lee_bound(
	s0_probs,
	s1_probs,
	y0_dists=None,
	y1_dists=None,
	# optional,
	y0_probs=None,
	y1_probs=None,
	y0_vals=None,
	y1_vals=None,
	m=1000,
):
	"""
	Helper function to compute semi-analytical Lee Bounds.


	Unlike dual bounds, this function is not at all
	robust to model misspecification. The estimand is 

	:math:`E[Y(1) - Y(0) | S(0) = S(1) = 1]`

	where :math:`Y(1), Y(0)` are potential outcomes and 
	:math:`S(1), S(0)` are post-treatment selection events.

	Parameters
	----------
	s0_probs : np.array
		n-length array where s0_probs[i] = P(S(0) = 1 | Xi)
	s1_probs : np.array
		n-length array where s1_probs[i] = P(S(1) = 1 | Xi)
	y0_dists : np.array
		batched scipy distribution of shape (n,) where the ith
		distribution is the conditional law of Yi(0) | S(0) = 1, Xi
	y1_dists : np.array
		batched scipy distribution of shape (n,) where the ith
		distribution is the conditional law of Yi(1) | S(1) = 1, Xi
	y0_vals : np.array
		nvals0-length array of values y0 can take.
	y0_probs : np.array
		(n, nvals0)-length array where
		y0_probs[i, j] = P(Y(0) = yvals0[j] | S(0) = 1, Xi)
	y1_vals : np.array
		(n,nvals1) array of values y1 can take.
	y1_probs : np.array
		(n, nvals1) array where
		y0_probs[i, j] = P(Y(1) = yvals1[j] | S(1) = 1, Xi)
	m : int
		Number of quantile discretizations to use when computing CVAR.
		m = 1000 (default) is more than sufficient.

	Returns
	-------
	agg_bounds : np.array
		(2,)-length array of lower and upper bound. This integrates 
		across all n y0_dists/y1_dists, etc.
	cond_bounds : np.array
		(2, n)-length array where bounds[0,i] is the ith lower bound
		and bounds[1,i] is the ith upper bound conditional on 
		:math:`X_i`.
	"""
	# Parse arguments
	n = s0_probs.shape[0]
	if y0_dists is None:
		y0_dists = BatchedCategorical(
			vals=y0_vals, probs=y0_probs
		)
	if y1_dists is None:
		y1_dists = BatchedCategorical(
			vals=y1_vals, probs=y1_probs
		)
	
	# always-takers share
	alphas = s0_probs / s1_probs
	if np.any(alphas > 1):
		raise ValueError(f"Monotonicity is violated for indices of alphas={alphas[alphas > 1]}")

	# compute E[Y(1) | Y(1) >= Q(alpha), S(1)=1]
	# (or <= depending on the value of lower)
	cvars_lower = compute_cvar(y1_dists, n, alpha=alphas, lower=True, m=m)
	cvars_upper = compute_cvar(y1_dists, n, alpha=1-alphas, lower=False, m=m)
	y0ms = y0_dists.mean()
	cond_bounds = np.stack([cvars_lower - y0ms, cvars_upper - y0ms], axis=0) # per-X bounds
	agg_bounds = np.mean(cond_bounds * s0_probs, axis=1) / np.mean(s0_probs) # aggregated bounds
	return agg_bounds, cond_bounds

def _lee_bound_no_covariates(
	y: np.array, W: np.array, S: np.array, pis: np.array
):
	"""
	Helper function which gets called by the public variant.
	"""
	# compute P(S | W)
	s0_prob = np.array([np.mean(S[W == 0])])
	s1_prob = np.array([np.mean(S[W == 1])])
	s1_prob = np.maximum(s0_prob, s1_prob)

	# compute P(Y(0))
	flags0 = (W == 0) & (S == 1)
	y0_vals = y[flags0]
	y0_probs = 1 / (1-pis[flags0]); y0_probs /= y0_probs.sum()
	inds0 = np.argsort(y0_vals)
	y0_vals = y0_vals[inds0]; y0_probs = y0_probs[inds0]

	# compute P(Y(1))
	flags1 = (W == 1) & (S == 1)
	y1_vals = y[flags1]
	y1_probs = 1 / (1-pis[flags1]); y1_probs /= y1_probs.sum()
	inds1 = np.argsort(y1_vals)
	y1_vals = y1_vals[inds1]; y1_probs = y1_probs[inds1]

	# Construct argument list
	args = dict(
		s0_probs=s0_prob,
		s1_probs=s1_prob,
		y0_probs=y0_probs.reshape(1, -1),
		y0_vals=y0_vals.reshape(1, -1),
		y1_probs=y1_probs.reshape(1, -1),
		y1_vals=y1_vals.reshape(1, -1)
	)
	# compute lower, upper bounds
	return compute_analytical_lee_bound(**args)[1][:, 0]
	

def lee_bound_no_covariates(
	outcome: np.array, 
	treatment: np.array, 
	selections: np.array, 
	propensities: Optional[np.array]=None,
	clusters: Optional[np.array]=None,
	B: int=200,
	alpha: float=0.05,
	verbose=False,
):
	"""
	Computes plug-in Lee bounds without using covariates.

	Parameters
	----------
	outcome : np.array
		n-length array of outcomes (y)
	treatment : np.array
		n-length array of treatments (W).
	selections : np.array
		n-length array of selection indicators (S).
	propensities : np.array
		n-length array of propensity scores (pis).
		Default: all equal to treatment.mean().
	clusters : np.array
		Optional n-length array of clusters, so ``clusters[i] = j``
		indicates that observation i is in cluster j.
	B : int
		Number of bootstrap replications to compute standard errors.
		Defaults to 0 (no standard errors).
	alpha : float
		nominal Type I error level.
	verbose : bool
		Show progress bar while bootstrapping if verbose=True.

	Returns
	-------
	results : dict
		Dictionary containing up to three keys:

		- estimates: 2-length array of lower/upper estimates.
		- ses: 2-length array of lower/upper standard errors.
		- cis: 2-length array of lower/upper confidence intervals.
	"""
	# Infer propensities
	if propensities is None:
		propensities = np.ones(len(treatment)) * treatment.mean()
	# Create estimates
	estimates = _lee_bound_no_covariates(
		y=outcome,
		W=treatment,
		S=selections,
		pis=propensities,
	)
	if B == 0:
		return dict(estimates=estimates)

	# Compute bootstrapped SEs
	data = np.stack([outcome, treatment, propensities, selections], axis=1)
	func = lambda data: _lee_bound_no_covariates(
		y=data[:, 0],
		W=data[:, 1],
		pis=data[:, 2],
		S=data[:, 3],
	)
	ses = utilities.cluster_bootstrap_se(
		data=data,
		clusters=clusters,
		func=func,
		B=B,
		verbose=verbose,
	)[0]
	cis = estimates.copy()
	cis[0] -= stats.norm.ppf(1-alpha/2) * ses[0]
	cis[1] += stats.norm.ppf(1-alpha/2) * ses[1]
	return dict(
		estimates=estimates,
		ses=ses,
		cis=cis,
	)

def lee_delta_method_se(
	sbetas, skappas, sgammas
):			
	# estimate
	hat_beta = sbetas.mean()
	hat_kappa = skappas.mean()
	hat_gamma = sgammas.mean()
	hattheta = (hat_beta - hat_kappa) / hat_gamma
	# standard error
	hatSigma = np.cov(
		np.stack([sbetas, skappas, sgammas], axis=0)
	) # 3 x 3 cov matrix
	grad = np.array([
		1 / hat_gamma,
		- 1 / hat_gamma,
		-(hat_beta - hat_kappa) / (hat_gamma**2)
	])
	# estimate
	se = np.sqrt(grad @ hatSigma @ grad / len(sbetas))
	return hattheta, se

class LeeDualBounds(generic.DualBounds):
	"""
	Computes dual bounds on the ATE under selection bias.

	Precisely, this class bounds 

	:math:`E[Y(1) - Y(0) | S(0) = S(1) = 1]`

	where :math:`Y(1), Y(0)` are potential outcomes and 
	:math:`S(1), S(0)` are post-treatment selection events.
	These bounds assume monotonicity, i.e., 
	:math:`S(1) >= S(0)` a.s. (see Lee 2009).

	Parameters
	----------
	selections : np.array
		n-length array-like of binary selection indicators
	outcome : np.array | pd.Series
		n-length array of outcome measurements (Y).
	treatment : np.array | pd.Series
		n-length array of binary treatment (W).
	covariates : np.array | pd.Series
		(n, p)-shaped array of covariates (X).
	propensities : np.array | pd.Series
		n-length array-like of propensity scores :math:`P(W=1 | X)`. 
		If ``None``, will be estimated from the data.
	outcome_model : str | dist_reg.DistReg
		The model for estimating the law of :math:`Y | X, W`.
		Two options:

		- A str identifier, e.g., 'ridge', 'lasso', 'elasticnet', 'randomforest', 'knn'.
		- An object inheriting from ``dist_reg.DistReg``. 

		E.g., when ``outcome`` is continuous, the default is
		``outcome_model=dist_reg.CtsDistReg(model_type='ridge')``.
	propensity_model : str | sklearn classifier
		How to estimate the propensity scores if they are not provided.
		Two options:

		- A str identifier, e.g., 'ridge', 'lasso', 'elasticnet', 'randomforest', 'knn'.
		- An sklearn classifier, e.g., ``sklearn.linear_model.LogisticRegressionCV()``.
	selection_model : str | dist_reg.BinaryDistReg
		How to estimate the selection probabilities :math:`P(S =1 | W, X)`.
		Two options:

		- A str identifier, i.e., 'monotone_logistic', 'ridge', 'lasso'.
		- An object inheriting from ``dist_reg.BinaryDistReg``.

		The default is ``monotone_logistic``. 
	support : np.array
		Optional support of the outcome, if known and discrete.
		Defaults to ``None`` (inferred from the data).
	model_kwargs : dict
		Additional kwargs for the ``outcome_model``, e.g.,
		``feature_transform``. See 
		:class:`dualbounds.dist_reg.CtsDistReg` or 
		:class:`dualbounds.dist_reg.BinaryDistReg` for more kwargs.
	"""	

	def __init__(
		self,
		selections: Union[np.array, pd.Series],
		*args, 
		selection_model: Optional[Union[str, dist_reg.BinaryDistReg]]=None,
		**kwargs
	):
		# Main initialization
		kwargs['f'] = None
		super().__init__(*args, **kwargs)
		# Initialization for S
		self.selection_model = selection_model
		if isinstance(selections, pd.Series) or isinstance(selections, pd.DataFrame):
			selections = selections.values.reshape(self.n)
		self.S = utilities._binarize_variable(selections, var_name='selections')
	
	def _ensure_feasibility(
		self,
		i,
		nu0,
		nu1,
		lower,
		ymin,
		ymax,
		grid_size=10000,
	):
		"""
		ensures nu0 + nu1 <= fvals (if lower)
		or nu0 + nu1 >= fvals (if upper)
		by performing a gridsearch.

		Parameters
		----------
		nu1 : nvals1-length array
			note nu1[0] is the dual variable for when s1=1.

		Returns
		-------
		new_nu0 : np.array
			``nvals0`` length array of new dual vars. for Y(0)
		new_nu1 : np.array
			``nvals1`` length array of new dual vars. for Y(1)
		dx : float
			Constant which can be subtracted off to obtain valid 
			dual variables. I.e., new_nk = nuk - dx/2.
		"""
		y1_vals = self.y1_vals_adj[i]
		if y1_vals[0] != 0:
			raise ValueError(
				"Expected y1_vals[0] to equal zero; this should be the case when s1 = 0."
			)
		dxs = []
		new_yvals = np.linspace(ymin, ymax, grid_size)
		interp_nu = self.interp_fn(
			x=y1_vals[1:], y=nu1[1:], newx=new_yvals,
		)
		# only three options due to monotonicity
		# if lower: nu0[s0] + nu1[s1 * y1] <= s0 s1 y1
		# if not lower: nu0[s0] + nu1[s1 * y1] >= s0 s1 y1
		for s0, s1 in zip(
			[0, 0, 1], [0, 1, 1]
		):
			if s1 == 0:
				dxs.append(nu0[s0] + nu1[0])
			else:
				deltas = nu0[s0] + interp_nu - s0 * s1 * new_yvals
				if lower:
					dxs.append(np.max(deltas))
				else:
					dxs.append(np.min(deltas))

		# return valid dual variables
		if lower:
			dx = np.max(np.array(dxs))
		else:
			dx = np.min(np.array(dxs))
		return nu0 - dx/2, nu1 - dx/2, dx

	def compute_dual_variables(
		self,
		s0_probs,
		s1_probs,
		y1_dists=None,
		y1_vals=None,
		y1_probs=None,
		verbose=False,
		nvals=100,
		ymin=None,
		ymax=None,
		**kwargs,
	):
		"""
		Estimates dual variables using the outcome model.

		We generally recommend that the user call .fit() instead of 
		calling this function directly.

		Parameters
		----------
		s0_probs : np.array
			n-length array where s0_probs[i] = :math:`P(S_i(0) = 1 | X_i)`.
		s1_probs : np.array
			n-length array where s1_probs[i] = :math:`P(S_i(1) = 1 | X_i)`.
		y1_dists : list
			The ith distribution of y1_dists represents the conditional
			law of :math:`Y_i(1) | X_i, S_i(1) =1`. There are two input formats:

			- batched scipy distribution of shape (n,)
			- list of scipy dists whose shapes add up to n.

		y1_vals : np.array
			(n, nvals1)-length array where ``y1_vals[i]`` is the support 
			of :math:`Y_i(1)`.	Ignored if ``y1_dists`` is provided.
		y1_probs : np.array
			(n, nvals1)-length array where ``y1_probs[i, j]``
			is the estimated probability that :math:`Y_i(1)`
			equals ``y1_vals[i, j].``
		verbose : bool
			If True, prints progress reports.
		nvals : int
			Number of values to use when discretizing Y(1).
		ymin : float
			Minimum value of Y(1) to use numerically.
		ymax : float
			Maximum value of Y(1) to use numerically.
		kwargs : dict
			kwargs for _ensure_feasibility method.
			Includes ymin, ymax, grid_size.
		"""
		# Constants for solver
		if verbose:
			print("Estimating optimal dual variables.")
		n = s0_probs.shape[0]
		self.nvals0 = 2 # because S is binary
		if self.discrete:
			self.nvals1 = len(self.support) + 1
		else:
			self.nvals1 = nvals
		self.interp_fn = interpolation.linear_interpolate
		ymin = ymin if ymin is not None else self.y.min()
		ymax = ymax if ymax is not None else self.y.max()

		# discretize if Y is continuous
		if y1_vals is None or y1_probs is None:
			# tolerance parameter
			min_quantile = min([np.min(s0_probs/s1_probs), np.min(1 - s0_probs/s1_probs)])
			# law of Y(1) | S(1) = 1, Xi
			y1_vals, y1_probs = self._discretize(
				y1_dists, 
				nvals=self.nvals1-1, 
				min_quantile=min_quantile/2,
				ninterp=1,
				ymin=ymin,
				ymax=ymax,
			)

		# ensure y1_vals, y1_probs are sorted
		self.y1_vals, self.y1_probs = utilities._sort_disc_dist(vals=y1_vals, probs=y1_probs)
		del y1_vals, y1_probs

		## Adjust to make it law of Y(1) S(1) | Xi
		# note: it is important that the value when S(1) = 0
		# is the first value on the second axis
		# so that the monotonicity constraint is correct.
		self.y1_vals_adj = np.concatenate(
			[np.zeros((n, 1)), self.y1_vals], axis=1
		)
		self.y1_probs_adj = np.concatenate(
			[
				1 - s1_probs.reshape(-1, 1),
				s1_probs.reshape(-1, 1) * self.y1_probs
			], 
			axis=1
		)

		# useful constants
		s0_vals = np.array([0, 1]).reshape(-1, 1)

		# Initialize results
		self.n = self.y1_vals.shape[0]
		self.nu0s = np.zeros((2, n, self.nvals0)) # first dimension = [lower, upper]
		self.nu1s = np.zeros((2, n, self.nvals1))
		self.hatnu0s = np.zeros((2, self.n))
		self.hatnu1s = np.zeros((2, self.n)) 
		# estimated cond means of nu0s, nu1s
		self.c0s = np.zeros((2, self.n))
		self.c1s = np.zeros((2, self.n))
		# objective values
		self.objvals = np.zeros((2, self.n))
		self.dxs = np.zeros((2, self.n))
		# loop through
		for i in utilities.vrange(self.n, verbose=verbose):
			# set parameter values
			fvals = (
				s0_vals * self.y1_vals_adj[i].reshape(1, -1)
			).astype(float)
			# below is the max val. used instead of inf to relax 
			# constraints. this provably has no effect compared
			# to using np.inf, but allows the use of ot instead
			# of cvxpy, leading to a substantial speedup.
			max_fval = np.abs(fvals).max() * 1e7
			# helpful concatenation
			probs0 = np.array([1 - s0_probs[i], s0_probs[i]])
			# solve 
			for lower in [1, 0]:
				# Relax constraints due to monotonicity
				fvals[1][0] = max_fval if lower else -max_fval
				nu0x, nu1x, objval = self._solve_single_instance(
					i=i,
					probs0=probs0,
					probs1=self.y1_probs_adj[i],
					y0_vals=np.array([0,1]),
					y1_vals=self.y1_vals_adj[i],
					fvals=fvals,
					not_binding=np.zeros(fvals.shape).astype(bool),
					lower=lower
				)
				self.objvals[1 - lower, i] = objval
				if not self.discrete:
					nu0x, nu1x, dx = self._ensure_feasibility(
						i=i, nu0=nu0x, nu1=nu1x,
						lower=lower, ymin=ymin, ymax=ymax,
						**kwargs,
					)
				else:
					dx = 0

				# Save intermediate quantities
				self.nu0s[1 - lower, i] = nu0x
				self.nu1s[1 - lower, i] = nu1x
				self.c0s[1 - lower, i] = nu0x @ probs0
				self.c1s[1 - lower, i] = nu1x @ self.y1_probs_adj[i]
				self.dxs[1 - lower, i] = dx

		self._compute_realized_dual_variables(y=self.y, S=self.S)

	def _compute_realized_dual_variables(self, y=None, S=None):
		y = self.y if y is None else y
		S = self.S if S is None else S
		### Compute realized hatnu1s/hatnu0s
		self.hatnu0s = np.zeros((2, self.n))
		self.hatnu1s = np.zeros((2, self.n))
		for i in range(self.n):
			for lower in [0, 1]:
				nu0x = self.nu0s[1-lower, i]
				nu1x = self.nu1s[1-lower, i]
				# Set values
				self.hatnu0s[1 - lower, i] = nu0x[S[i]]
				if S[i] == 0:
					self.hatnu1s[1 - lower, i] = nu1x[0]
				if not self.discrete and S[i] == 1:
					self.hatnu1s[1 - lower, i] = self.interp_fn(
						x=self.y1_vals_adj[i][1:], y=nu1x[1:], newx=y[i],
					)[0]
				if self.discrete and S[i] == 1:
					j = np.argmin(np.abs(self.y1_vals_adj[i][1:] - y[i]))
					self.hatnu1s[1 - lower, i] = nu1x[j+1]

	def cross_fit(
		self,
		nfolds: int=5,
		suppress_warning: bool=False,
		verbose: bool=True,
	):
		"""
		Cross-fits the outcome and selection models.

		Parameters
		----------
		nfolds : int
			Number of folds to use in cross-fitting.
		suppress_warning : bool
			If True, suppresses the warning about manual crossfitting.
		verbose : bool
			If True, prints progress reports.

		Returns
		-------
		s0_probs : np.array
			n-length array where s0_probs[i] = :math:`P(S_i(0) = 1 | X_i)`.
		s1_probs : np.array
			n-length array where s1_probs[i] = :math:`P(S_i(1) = 1 | X_i)`.
		y0_dists : np.array
			list of batched scipy distributions whose shapes sum to n.
			the ith dist. is the conditional law of :math:`Y_i(0) | S_i(0) = 1, X_i`.
		y1_dists : list
			list of batched scipy distributions whose shapes sum to n.
			the ith dist. is the conditional law of :math:`Y_i(1) | S_i(1) = 1, X_i`.
		"""
		# estimate selection probs
		if self.s0_probs is None or self.s1_probs is None:
			if self.selection_model is None:
				self.selection_model = 'monotone_logistic'
			self.selection_model = generic.get_default_model(
				outcome_model=self.selection_model, 
				# the following args are ignored if selection_model already
				# inherits from dist_reg.DistReg class
				support=set([0,1]), 
				discrete=True,
				monotonicity=True, 
				how_transform='intercept',
			)
			if verbose:
				print("Cross-fitting the selection model.")
			sout = dist_reg.cross_fit_predictions(
				W=self.W, X=self.X, y=self.S, 
				nfolds=nfolds, 
				model=self.selection_model,
				verbose=verbose,
				probs_only=True,
			)
			counterfactuals, self.selection_model_fits, self.S_oos_preds = sout
			self.s0_probs, self.s1_probs = counterfactuals
		elif not suppress_warning:
			warnings.warn(
				generic.CROSSFIT_WARNING.replace("y0_", "s0_").replace("y1_", "s1_")
			)

		# Estimate outcome model
		if self.y0_dists is None or self.y1_dists is None:
			self.outcome_model = generic.get_default_model(
				discrete=self.discrete, support=self.support, outcome_model=self.outcome_model,
			)
			if verbose:
				print("Cross-fitting the outcome model.")
			yout = dist_reg.cross_fit_predictions(
				W=self.W, X=self.X, S=self.S, y=self.y, 
				nfolds=nfolds, 
				model=self.outcome_model,
				verbose=verbose,
			)
			counterfactuals, self.model_fits, self.oos_dist_preds = yout
			self.y0_dists, self.y1_dists = counterfactuals
		elif not suppress_warning:
			warnings.warn(generic.CROSSFIT_WARNING)

		# return
		return self.s0_probs, self.s1_probs, self.y0_dists, self.y1_dists


	def fit(
		self,
		nfolds=5,
		alpha=0.05,
		aipw=True,
		s0_probs=None,
		s1_probs=None,
		y0_dists=None,
		y1_dists=None,
		suppress_warning=False,
		verbose=True,
		**solve_kwargs,
	):
		"""
		Main function which (1) performs cross-fitting, (2) computes 
		optimal dual variables, and (3) computes final dual bounds.

		Parameters
		----------
		nfolds : int
			Number of folds to use when cross-fitting. Defaults to 5,
		alpha : float
			Nominal coverage level. Defaults to 0.05.
		aipw : bool
			If true, returns AIPW estimator.
		s0_probs : np.array
			Optional n-length array where s0_probs[i] = 
			:math:`P(S_i(0) = 1 | X_i)`.
			If not provided, will be estimated from the data.
		s1_probs : np.array
			Optional n-length array where s1_probs[i] = 
			:math:`P(S_i(1) = 1 | X_i)`.
			If not provided, will be estimated from the data.
		y0_dists : np.array
			Optional list of batched scipy distributions whose shapes sum to n.
			the ith dist. is the conditional law of 
			:math:`Y_i(0) | S_i(0) = 1, X_i`.
			If not provided, will be estimated from the data.
		y1_dists : np.array
			Optional list of batched scipy distributions whose shapes sum to n.
			The ith dist. is the conditional law of 
			:math:`Y_i(1) | S_i(1) = 1, X_i`.
			If not provided, will be estimated from the data.
		suppress_warning : bool
			If True, suppresses warning about cross-fitting.
		verbose : bool
			If True, gives occasional progress reports..
		solve_kwargs : dict
			kwargs to self.compute_dual_variables(), 
			e.g., ``verbose``, ``nvals``, ``grid_size``

		Returns
		-------
		self : object
		"""
		# Save data
		self.s0_probs, self.s1_probs = s0_probs, s1_probs
		self.y0_dists, self.y1_dists = y0_dists, y1_dists

		# if pis not supplied: will use cross-fitting
		if self.pis is None:
			self.fit_propensity_scores(verbose=verbose, nfolds=nfolds)

		# fit outcome models using cross-fitting
		self.cross_fit(
			verbose=verbose, nfolds=nfolds, suppress_warning=suppress_warning,
		)

		# compute dual variables
		self.compute_dual_variables(
			s0_probs=self.s0_probs,
			s1_probs=self.s1_probs,
			y1_dists=self.y1_dists,
			ymin=self.y.min(),
			ymax=self.y.max(),
			verbose=verbose,
			**solve_kwargs,
		)

		# compute dual bounds
		self._compute_final_bounds(aipw=aipw, alpha=alpha)
		return self

	def _compute_final_bounds(self, aipw=True, alpha=0.05):
		"""
		Computes final bounds based in (A)IPW summands,
		using the delta method for E[Y(1) - Y(0) | S(1) = S(0) = 1].
		Uses the bootstrap for clustered standard errors.
		"""
		self._compute_ipw_summands()
		summands = self.aipw_summands if aipw else self.ipw_summands
		self._compute_cond_means()
		self.y0s0_cond_means = self.mu0 * self.s0_probs
		self.s_probs = self.s0_probs.copy()
		self.s_probs[self.W == 1] = self.s1_probs[self.W == 1]
		ests = []
		ses = []
		bounds = []
		scale = stats.norm.ppf(1-alpha/2)
		# kappa = E[Y(0) S(0)]
		skappas = (1 - self.W) * (self.y * self.S - self.y0s0_cond_means) 
		skappas = skappas / (1-self.pis)
		skappas += self.y0s0_cond_means
		self.skappas = skappas
		# gamma = E[S(0)]
		sgammas = (1-self.W) * (self.S - self.s0_probs) / (1-self.pis)
		sgammas += self.s0_probs
		self.sgammas = sgammas
		if self.clusters is None:
			for lower in [1, 0]:
				# beta = part. identifiable component E[Y(1) S(0)]
				sbetas = summands[1-lower]
				hattheta, se = lee_delta_method_se(
					sbetas=sbetas, skappas=skappas, sgammas=sgammas,
				)
				ests.append(hattheta)
				ses.append(se)
				if lower:
					bounds.append(hattheta - scale * se)
				else:
					bounds.append(hattheta + scale * se)

			self.estimates = np.array(ests)
			self.ses = np.array(ses)
			self.cis = np.array(bounds)
			return dict(
				estimates=self.estimates,
				ses=self.ses,
				cis=self.cis
			)
		else:
			self.estimates, self.ses, self.cis = _delta_bootstrap_ses(
				h=lambda fval, z0, z1: (fval - z0[0]) / z0[1],
				summands=summands,
				z0summands=np.stack([skappas, sgammas], axis=1),
				z1summands=np.zeros((self.n, 1)),
				clusters=self.clusters,
				alpha=alpha,
			)
			# Return
			return dict(
				estimates=self.estimates,
				ses=self.ses,
				cis=self.cis,
			)


	def summary(self, minval=-np.inf, maxval=np.inf):
		"""
		Prints a summary of main results from the class.

		Parameters
		----------
		minval : float
			Analytical lower bound on estimand used to clip results. 
			Defaults to -np.inf.
		maxval : float
			Analytical upper bound on estimand used to clip results.
			Defaults to np.inf.
		"""
		print("___________________Inference_____________________")
		print(self.results(minval=minval, maxval=maxval))
		print()
		print("________________Selection model__________________")
		sumstats = dist_reg._evaluate_model_predictions(
			y=self.S, haty=self.s_probs
		)
		print(sumstats)
		print()
		print("_________________Outcome model___________________")
		self._compute_oos_resids()
		sumstats = dist_reg._evaluate_model_predictions(
			y=self.y[self.S == 1], haty=self.oos_preds[self.S == 1],
		)
		print(sumstats)
		print()
		print("________________Treatment model__________________")
		sumstats = dist_reg._evaluate_model_predictions(
			y=self.W, haty=self.pis
		)
		print(sumstats)
		print()


import time
import numpy as np
import scipy as sp
from scipy import stats
import unittest
import pytest
import os
import sys
try:
	from . import context
	from .context import dualbounds as db
# For profiling
except ImportError:
	import context
	from context import dualbounds as db

from dualbounds import utilities

class TestUtils(unittest.TestCase):
	"""
	tests parse_dist
	"""
	def test_adjust_support_size(self):
		# A test where we know the result
		vals = np.array([0, 1, 2, 3, 3.1, 4.01, 5.05])
		probs = np.array([5., 1, 1, 1, 1, 1, 1])
		probs /= probs.sum()
		newvals, newprobs = utilities._adjust_support_size_unbatched(
			vals, probs, new_nvals=5, ymin=None, ymax=None
		)
		evals = np.array([0, 1.5, 3.05, 4.01, 5.05])
		eprobs = np.array([5., 2, 2, 1, 1])
		eprobs /= eprobs.sum()
		for out, expected in zip([newvals, newprobs], [evals, eprobs]):
			np.testing.assert_array_almost_equal(
				out, expected, decimal=4, 
				err_msg="greedy _adjust_support_size gives incorrect result"
			)

		# Test results where we don't know the answer
		n = 10
		new_nvals = 5
		for nvals in [new_nvals - 1, new_nvals, new_nvals + 20]:
			vals = np.sort(np.random.randn(n, nvals), axis=1)
			probs = np.random.uniform(size=(n, nvals))
			probs /= probs.sum(axis=1).reshape(-1, 1)
			newvals, newprobs = utilities.adjust_support_size(
				vals, probs, new_nvals=new_nvals, ymin=-1, ymax=1
			)
			self.assertTrue(
				newvals.shape == (n, new_nvals),
				f"Adjusted dist has wrong shape: expected n={n}, nvals={new_nvals}"
			)
			np.testing.assert_array_almost_equal(
				np.sum(vals * probs, axis=1),
				np.sum(newvals * newprobs, axis=1),
				decimal=6,
				err_msg="Mean of original/adjusted distributions are different"
			)


	def test_parse_dist_mu_sd(self):
		"""
		Ensures returns dists with correct mean/sd.
		"""
		n = 20
		mu = np.random.randn(n)
		sd = np.random.uniform(n)
		for dist, kwargs in zip(
			['gamma', 'invchi2', 'tdist', 'expon'],
			[dict(a=5), dict(df=3), dict(df=6), dict()],
		):
			out = utilities.parse_dist(
				dist, mu=mu, sd=sd, **kwargs
			)
			# assert equality
			np.testing.assert_array_almost_equal(
				out.mean(), mu, 
				decimal=3,
				err_msg=f"parse_dist mean is wrong for dist={dist}, {kwargs}"
			)
			# assert equality
			np.testing.assert_array_almost_equal(
				out.std(), sd, 
				decimal=3,
				err_msg=f"parse_dist sd is wrong for dist={dist}, {kwargs}"
			)

		# test uniform
		mu = 0
		sd = np.sqrt(4 / 12)
		out = utilities.parse_dist('uniform', mu=0, sd=sd)
		np.testing.assert_array_almost_equal(
			np.array(out.support()),
			np.array([-1, 1]),
			decimal=3,
			err_msg=f"parse_dist returns wrong support for uniform with mu=0, sd={sd}"
		)

	def test_se_computation(self):
		np.random.seed(123)
		reps = 100
		n = 300
		n_clusters = 150
		for d, func in zip(
			[1, 1, 1, 2], [None, lambda x: x[0], lambda x: x[0]**2, lambda x: x[0] * x[1]**2]
		):

			# True means, variances and parameters
			mus = np.random.randn(2, d) + 5
			sigmas = np.random.uniform(0.5, 1, size=(2, d)) 
			sigmas += np.array([0, 10]).reshape(2, 1)
			clusters = np.random.choice(n_clusters, size=n, replace=True)
			# Initialize storage
			ests = np.zeros((reps, 2))
			ses = np.zeros((reps, 2))
			for r in range(reps):
				# 100% within-cluster correlation
				samples = np.random.randn(2, n_clusters, d) * sigmas.reshape(2, 1, d)
				samples += mus.reshape(2, 1, d)
				samples = samples[:, clusters]
				# bootstrap + ses
				ests[r], ses[r], _ = utilities.compute_est_bounds(
					summands=samples,
					func=func,
					clusters=clusters,
					B=100,
				)

			# Compare true and estimated ses
			true_ses = ests.std(axis=0)
			bs_ses = ses.mean(axis=0)
			self.assertTrue(
				np.all(true_ses / bs_ses <= 1.2),
				f"true_ses={true_ses} and bs_ses={bs_ses} are different"
			)
			self.assertTrue(
				np.all(bs_ses / true_ses <= 1.2),
				f"true_ses={true_ses} and bs_ses={bs_ses} are different"
			)

	def test_weighted_quantile(self):
		np.random.seed(123)
		n = 101
		# Data and desired quantiles
		x = np.random.randn(n)
		probs = np.ones(n) / n
		qs = np.linspace(0, 1, n-1)
		# Check that weighted_quantile agrees with np.quantile
		out = utilities.weighted_quantile(x, probs, qs)
		expected = np.quantile(x, qs)
		np.testing.assert_array_almost_equal(
			out,
			expected,
			decimal=5,
			err_msg=f"weighted_quantile with equal weights differs from np.quantile"
		)
		# Check weighted variant
		n = 1001
		x = np.linspace(0, 1, n)
		probs = x**2 / np.sum(x**2)
		out = utilities.weighted_quantile(x, probs, 0.5)
		expected = np.median(
			np.random.choice(x, p=probs, size=100000, replace=True)
		)
		np.testing.assert_array_almost_equal(
			out,
			expected,
			decimal=2,
			err_msg=f"weighted_quantile median from sampling approach"
		)


import time
import numpy as np
import scipy as sp
from scipy import stats
import unittest
import pytest
import os
import sys
try:
	from . import context
	from .context import dualbounds as db
# For profiling
except ImportError:
	import context
	from context import dualbounds as db

from dualbounds import bootstrap, utilities, gen_data

class TestMultiplierBootstrap(unittest.TestCase):
	"""
	tests multiplier bootstrap methods
	"""
	def test_multiplier_bootstrap_simple(self):
		# Parameters
		np.random.seed(1234)
		n, reps = 500, 200
		for d, eps_dist in zip(
			[1, 2, 5, 20],
			['laplace', 'gaussian', 'tdist', 'tdist']
		):
			alpha = 0.2
			mu = np.random.randn(d)
			sigmas = np.random.uniform(size=d)
			# Loop through and compute
			cis_upper = np.zeros(reps)
			cis_lower = np.zeros(reps)
			for r in range(reps):
				samples = utilities.parse_dist(
					eps_dist, mu=mu, sd=sigmas,
				).rvs(size=(n,d))
				# Test the ability to handle zero-variance columns
				if d > 1:
					samples[:, -1] = mu[-1]

				# Compute bounds
				_, ciu = bootstrap.multiplier_bootstrap(
					samples=samples, alpha=alpha, B=100
				)
				cis_upper[r] = ciu
				_, cil = bootstrap.multiplier_bootstrap(
					samples=samples, alpha=alpha, B=100, param='min'
				)
				cis_lower[r] = cil


			error_upper = np.mean(cis_upper > np.max(mu))
			error_lower = np.mean(cis_lower < np.min(mu))
			for error, name in zip([error_lower, error_upper], ['lower', 'upper']):
				self.assertTrue(
					error <= alpha + 3 * np.sqrt(alpha * (1-alpha) / reps),
					f"Multiplier bootstrap error={error} > alpha={alpha} for {name} CI with d={d}."
				)

	def test_db_multiplier_cluster_bootstrap_agree(self):
		# Fit two dualbounds objects
		data = db.gen_data.gen_regression_data(n=200, p=5, sample_seed=123)
		db_objects = []
		for Y_model in ['ridge', 'knn']:
			gdb = db.generic.DualBounds(
				outcome=data['y'], 
				covariates=data['X'], 
				treatment=data['W'], 
				propensities=data['pis'],
				outcome_model=Y_model,
				f=lambda y0, y1, x: (y1 < y0).astype(float)
			)
			gdb.fit(nfolds=2)
			db_objects.append(gdb)

		# Run mult bootstrap
		alpha = 0.1
		mbs_output = bootstrap.dualbound_multiplier_bootstrap(
			db_objects, alpha=alpha, B=10000
		)
		# Run cluster bootstrap
		cbs_output = bootstrap.dualbound_cluster_bootstrap(
			db_objects, alpha=alpha, B=10000
		)
		# Check that the results agree
		np.testing.assert_array_almost_equal(
			mbs_output.values,
			cbs_output.values,
			decimal=2,
			err_msg=f"Multiplier and cluster bootstrap do not agree."
		)


import time
import numpy as np
import scipy as sp
from scipy import stats
import unittest
import pytest
import os
import sys
try:
	from . import context
	from .context import dualbounds as db
# For profiling
except ImportError:
	import context
	from context import dualbounds as db

from dualbounds.interpolation import adaptive_interpolate

class TestInterpolation(unittest.TestCase):
	"""
	tests parse_dist
	"""
	def test_interpolation(self):
		# A test which should use linear interpolation
		n = 5
		x = np.arange(n)
		y = np.arange(n)
		newx = np.arange(20) - 10
		expected = newx.copy()
		out = adaptive_interpolate(x, y, newx)
		np.testing.assert_array_almost_equal(
			expected, out, decimal=5, err_msg="adaptive_interpolate fails for a linear function"
		)
		# A test which should use nearest-neighbor interpolation
		x = np.arange(n)
		y = x >= 2
		newx = np.array([-20, 1.4, 2.1, 20])
		expected = newx >= 2
		out = adaptive_interpolate(x, y, newx)
		np.testing.assert_array_almost_equal(
			expected, out, decimal=5, err_msg="adaptive_interpolate fails for indicator function"
		)


import time
import numpy as np
import scipy as sp
from scipy import stats
import unittest
import pytest
import os
import sys
try:
	from . import context
	from .context import dualbounds as db
# For profiling
except ImportError:
	import context
	from context import dualbounds as db

from dualbounds import varite, utilities, gen_data


class TestVarITE(unittest.TestCase):

	def test_varite_consistency(self):
		""" This tests consistent Var(ITE) estimation. """
		n = 3000
		for tau, r2 in zip([0, 5, -2], [0.9, 0.0, 0.99]):
			data = gen_data.gen_regression_data(
				n=n, p=5, r2=r2, tau=tau, eps_dist='bernoulli', sample_seed=123,
			)
			# Compute true upper bound
			lower, upper = varite.compute_analytical_varite_bound(
				n=n, 
				y0_dists=data['y0_dists'], y1_dists=data['y1_dists'],
				reps=10,
			)

			# Compute vdb
			vdb = db.varite.VarITEDualBounds(
				covariates=data['X'],
				outcome=data['y'],
				treatment=data['W'],
				propensities=data['pis'],
			)
			ests = vdb.fit(nfolds=3).estimates

			# test accuracy
			np.testing.assert_array_almost_equal(
				np.array([lower, upper]),
				ests,
				decimal=1,
				err_msg=f"Var(ITE) bounds are not consistent, n={n}"
			)

	def test_varite_delta_method_se(self):
		""" Tests that we correctly estimate the SE. """
		context._run_se_computation_test(
			dim=3,
			f=lambda x, y, z: x - (y-z)**2,
			arg_names=['sbetas', 'skappa1s', 'skappa0s'],
			testname='Var(ITE)',
			se_function=db.varite.varite_delta_method_se,
		)

	def test_no_error(self):
		""" This tests if the code runs without error for continuous Y. """
		for eps_dist in ['gaussian', 'bernoulli']:
			data = gen_data.gen_regression_data(
				n=200, p=10, eps_dist=eps_dist, sample_seed=123
			)
			vdb = db.varite.VarITEDualBounds(
				covariates=data['X'],
				outcome=data['y'],
				treatment=data['W'],
				propensities=data['pis'],
			)
			vdb.fit(nfolds=3).summary()


class TestVarCATE(unittest.TestCase):

	def test_varcate_delta_method_se(self):
		""" Tests that we correctly estimate the SE. """
		context._run_se_computation_test(
			dim=6,
			f=db.varcate._moments2varcate,
			arg_names=[
				'shxy1', 'shxy0', 'shx', 'sy1', 'sy0', 'shx2'
			],
			testname='Var(CATE)',
			se_function=db.varcate.varcate_delta_method_se,
		)

	def test_no_discrete_error(self):
		"""
		Just tests that the discrete case runs without errors.
		The LogisticCV solver is slow so we don't check consistency.
		"""
		data = db.gen_data.gen_lee_bound_data(
			n=200, p=5, eps_dist='bernoulli',
		)
		vdb = db.varcate.VarCATEDualBounds(
			covariates=data['X'], outcome=data['y'], treatment=data['W'], propensities=data['pis'],
		)
		vdb.fit(nfolds=3).summary()


	def test_varcate_consistency(self):
		for eps_dist, r2 in zip(
			['gaussian', 'expon'], [0.9, 0.5, 0.0]
		):
			data = db.gen_data.gen_lee_bound_data(
				n=300000, p=5, r2=r2, tau=2, dgp_seed=1, sample_seed=1,
				eps_dist=eps_dist,
				interactions=True,
			)
			expected = data['cates'].std()**2
			## Oracle
			vdb_oracle = db.varcate.VarCATEDualBounds(
				covariates=data['X'], outcome=data['y'], treatment=data['W'], propensities=data['pis'],
			)
			est_oracle = vdb_oracle.fit(
				y0_dists=data['y0_dists'], y1_dists=data['y1_dists'], 
				suppress_warning=True,
			).estimates[0]
			## Actual
			vdb = db.varcate.VarCATEDualBounds(
				covariates=data['X'], outcome=data['y'], treatment=data['W'], propensities=data['pis'],
				outcome_model=db.dist_reg.CtsDistReg(eps_dist='gaussian')
			)
			est_actual = vdb.fit(nfolds=3).estimates[0]
			for est, name in zip([est_oracle, est_actual], ['oracle', 'est']):
				np.testing.assert_array_almost_equal(
					est,
					expected,
					decimal=2,
					err_msg=f"VarCATE {name} is inaccurate (r2={r2}, eps_dist={eps_dist})"
				)

	def test_varcate_cluster_bootstrap(self):
		data = db.gen_data.gen_regression_data(n=300, p=10, sparsity=0.8, sample_seed=123)
		# Fit multiple varcate models
		vdbs = []
		outcome_models = ['lasso', db.dist_reg.CtsDistReg('knn', n_neighbors=50)]
		for outcome_model in outcome_models:
			vdb = db.varcate.VarCATEDualBounds(
				outcome=data['y'],
				treatment=data['W'],
				covariates=data['X'],
				propensities=data['pis'],
				outcome_model=outcome_model,
			).fit(verbose=False)
			vdbs.append(vdb)
		# Combine
		vdb_results = db.varcate.varcate_cluster_bootstrap(vdbs, verbose=False, alpha=0.05, B=1000)
		# Check estimate
		expected_estimate = np.max([x.estimates[0] for x in vdbs])
		estimate = vdb_results.loc['Estimate', 'Lower'].item()
		self.assertTrue(
			estimate==expected_estimate,
			msg=f"Combined estimate={estimate} is different than the max estimate {expected_estimate}"
		)
		# Check CI
		expected_ci = np.max([x.cis[0] for x in vdbs])
		ci = vdb_results.loc['Conf. Int.', 'Lower'].item()
		ci_tol = vdbs[np.argmin([x.cis[0] for x in vdbs])].ses[0]
		self.assertTrue(
			ci >= expected_ci - 2 * ci_tol,
			msg=f"Combined CI={ci} <= max lower_ci={expected_ci} plus tolerance {ci_tol}"
		)
		self.assertTrue(
			ci <= expected_ci,
			msg=f"Combined CI={ci} is somehow sharper than max lower_ci={expected_ci}"
		)

		# for expected, output, tol, decimal, name in zip(
		# 	[expected_estimate, expected_ci], 
		# 	[estimate, ci],
		# 	[0, ci_tol],
		# ):

	def test_calibrated_varcate_db(self):
		outcome_models = ['ridge', 'knn']
		for eps_dist, outcome_models in zip(
			['gaussian', 'bernoulli'],
			[['ridge', 'knn'], ['knn', db.dist_reg.BinaryDistReg('rf', min_samples_leaf=50, n_estimators=3)]],
		):
			data = db.gen_data.gen_regression_data(n=300, p=5, sample_seed=123, eps_dist=eps_dist)
			# fit
			vdb = db.varcate.CalibratedVarCATEDualBounds(
				outcome=data['y'],
				treatment=data['W'],
				covariates=data['X'],
				propensities=data['pis'],
				outcome_model=outcome_models,
			).fit(nfolds=3, verbose=False).summary()

if __name__ == "__main__":
	# Run all tests---useful if using cprofilev
	basename = os.path.basename(os.path.abspath(__file__))
	if sys.argv[0] == f'test/{basename}':
		time0 = time.time()
		context.run_all_tests([TestVarITE(), TestVarCATE()])
		elapsed = np.around(time.time() - time0, 2)
		print(f"Finished running all tests at time={elapsed}")

	# Else let unittest handle this for us
	else:
		unittest.main()


import numpy as np
import pandas as pd
from scipy import stats
from .utilities import vrange, cluster_bootstrap_se, itemize
from .generic import DualBounds
from .delta import DeltaDualBounds

def multiplier_bootstrap(
	samples: np.array, 
	alpha: float,
	B: int=1000, 
	maxarrsize: int=int(1e10),
	param: str='max',
	verbose: int=False,
):
	"""
	Computes multiplier bootstrap lower confidence bounds. 

	Precisely, computes a lower confidence bound on
	:math:`\max(\mu_1, \dots, \mu_d)`,
	where :math:`\mu_i` is the mean of ``samples[i]``.

	Parameters
	----------
	samples : np.array
		(n,d)-shaped array where samples[i]
		is i.i.d. with mean :math:`\mu_i`.
	alpha : float
		Nominal error control level.
	B : int
		Number of bootstrap replications
	maxarrsize : float
		Maximum size of an array; used to save memory.
	param : str
		
		- If param='max', computes a lower confidence bound on :math:`\max(\mu_1, \dots, \mu_d)`.
		- Else, computes an upper confidence bound on :math:`\min(\mu_1, \dots, \mu_d)`,

	verbose : bool
		If True, shows a progress bar. Only useful
		if ``samples`` is a very large matrix.

	Returns
	-------
	estimate : float
		Estimate of :math:`\max(\mu_1, \dots, \mu_d)`.
	ci : float
		Lower confidence bound on :math:`\max(\mu_1, \dots, \mu_d)`.
	"""
	if param != 'max':
		estimate, ci = multiplier_bootstrap(
			samples=-1 * samples, 
			alpha=alpha, 
			B=B, 
			maxarrsize=maxarrsize,
			param='max',
		)
		return -estimate, -ci

	hatmu = samples.mean(axis=0)
	hatsigma = samples.std(axis=0)
	# This is important: if hatsigma \approx 0,
	# treat it as zero to ensure numerical stability when dividing.
	# Otherwise this should have zero effect.
	hatsigma = np.around(hatsigma, 10) 
	if np.all(hatsigma == 0):
		return np.max(hatmu), np.max(hatmu)
	if np.any(hatsigma == 0):
		# use noise-free columns as a deterministic lower bound
		min_val = np.max(hatmu[hatsigma == 0])
		samples = samples[:, hatsigma != 0]
		hatmu = hatmu[hatsigma != 0]
		hatsigma = hatsigma[hatsigma != 0]
	else:
		min_val = - np.inf

	# Centered statistics
	Tbs = []

	# Determine batch size
	n, d = samples.shape
	batchsize = min(B, max(1, int(maxarrsize / (n * d))))
	n_batches = int(np.ceil(B / batchsize))
	# Loop and compute bootstrap
	for b in vrange(n_batches, verbose=verbose):
		W = np.random.randn(n, batchsize, 1)
		sw = np.sum(
			W * (samples - hatmu).reshape(n, 1, d),
			axis=0
		) / np.sqrt(n)
		Tbs.append(np.max(sw / hatsigma, axis=1))
	Tbs = np.concatenate(Tbs)
	# Compute quantile and upper CI
	quantile = np.quantile(Tbs, 1-alpha)
	estimate = max(np.max(hatmu), min_val)
	ci = max(
		np.max(hatmu - quantile * hatsigma / np.sqrt(n)), 
		min_val
	)
	return estimate, ci

def dualbound_cluster_bootstrap(
	db_objects: list[DualBounds],
	aipw: bool=True,
	alpha: float=0.05,
	B: int=1000,
):
	"""
	Combines evidence across multiple DualBounds classes
	using a (clustered) bootstrap.

	Parameters
	----------
	db_objects : list
		A list of fit DualBounds classes.
	aipw : bool
		If True, uses AIPW estimators to reduce variance
		(highly recommended).
	alpha : float
		Nominal level, between 0 and 1.
	B : int
		Number of bootstrap replications.
	"""
	# Learn dimensionality
	K = len(db_objects)
	n = db_objects[0].aipw_summands.shape[1]
	# Original estimates and SEs
	estimates = np.stack([x.estimates for x in db_objects], axis=-1) # 2 X K
	ses = np.stack([x.ses for x in db_objects], axis=-1) # 2 X K
	# Combine all of the summands appropriately
	if aipw:
		summands = np.stack([x.aipw_summands for x in db_objects], axis=-1) # 2 X n x K
	else:
		summands = np.stack([x.ipw_summands for x in db_objects], axis=-1)
	if isinstance(db_objects[0], DeltaDualBounds):
		z0summands = np.stack([x.z0summands for x in db_objects], axis=-1) # n x d0 x K
		z1summands = np.stack([x.z1summands for x in db_objects], axis=-1) # n x d1 x K
		# Extract functions
		h = db_objects[0].h
	else:
		z0summands = np.zeros((n, 1, K))
		z1summands = np.zeros((n, 1, K))
		# Dummy versions of the DeltaDualBounds functions
		h = lambda fval, z0, z1: fval

	d0 = z0summands.shape[1]
	def compute_estimate(data):
		hatmu = data.mean(axis=0) # (1 + d0 + d1) x K
		return np.array([
			itemize(h(fval=hatmu[0, k], z0=hatmu[1:(d0+1), k], z1=hatmu[(d0+1):, k]))
			for k in range(K)
		])
		
	# Bootstrap
	combined_estimates = np.zeros(2)
	cis = np.zeros(2)
	#abe = np.zeros((2, B, K))
	#cbe = np.zeros((2, B, K))
	for lower in [0, 1]:
		samples = np.concatenate(
			[summands[1-lower].reshape(n, 1, K), z0summands, z1summands], axis=1
		) # n x (1 + d0 + d1) x K
		# Bootstrapped estimators
		_, bs_estimators = cluster_bootstrap_se(
			data=samples, 
			clusters=db_objects[0].clusters,
			func=compute_estimate,
			B=B,
			verbose=False,
		) # B x K
		# Centerd and standardized variant
		centered = (bs_estimators - estimates[1-lower]) / ses[1-lower]
		# Compute final estimates and bounds
		if lower == 1:
			hatq = np.quantile(centered.max(axis=-1), 1-alpha/2)
			combined_estimates[1-lower] = np.max(estimates[1-lower])
			cis[1-lower] = np.max(estimates[1-lower] - hatq * ses[1-lower])
		else:
			hatq = np.quantile(centered.min(axis=-1), alpha/2)
			combined_estimates[1-lower] = np.min(estimates[1-lower])
			cis[1-lower] = np.min(estimates[1-lower] - hatq * ses[1-lower])
	# Return
	return pd.DataFrame(
		np.stack([combined_estimates, cis], axis=0),
		index=['Estimate', 'Conf. Int.'],
		columns=['Lower', 'Upper'],
	)



def dualbound_multiplier_bootstrap(
	db_objects: list[DualBounds], 
	aipw: bool=True,
	alpha: float=0.05,
	**kwargs,
) -> pd.DataFrame:
	"""
	Combines evidence across multiple DualBounds classes
	using the multiplier bootstrap.

	Parameters
	----------
	db_objects : list
		A list of fit DualBounds classes.
	aipw : bool
		If True, uses AIPW estimators to reduce variance
		(highly recommended).
	alpha : float
		Nominal level, between 0 and 1.
	kwargs : dict
		kwargs for dualbounds.bootstrap.multiplier_bootstrap.
	Returns
	-------
	result : pd.DataFrame
		dataframe of results.
	"""
	# Fetch summands
	if aipw:
		summands = [x.aipw_summands for x in db_objects]
	else:
		summands = [x.ipw_summands for x in db_objects]

	# Separate lower/upper summands
	lower_summands = np.stack([x[0, :] for x in summands], axis=1)
	upper_summands = np.stack([x[1, :] for x in summands], axis=1)

	# Compute lower/upper CI
	lower_est, lower_ci = multiplier_bootstrap(
		samples=lower_summands,
		param='max',
		alpha=alpha/2,
		**kwargs
	)
	upper_est, upper_ci = multiplier_bootstrap(
		samples=upper_summands,
		param='min',
		alpha=alpha/2,
		**kwargs
	)
	# Return
	estimates = np.array([lower_est, upper_est])
	cis = np.array([lower_ci, upper_ci])
	return pd.DataFrame(
		np.stack([estimates, cis], axis=0),
		index=['Estimate', 'Conf. Int.'],
		columns=['Lower', 'Upper'],
	)


import copy
import warnings
import numpy as np
import pandas as pd
import ot
from scipy import stats
from . import utilities
from .generic import DualBounds
from typing import Optional

def _delta_bootstrap_ses(
	h: callable,
	summands: np.array,
	z1summands: np.array,
	z0summands: np.array,
	alpha: float,
	clusters: Optional[np.array]=None,
	B: int=1000,
):
	"""
	Uses the bootstrap to compute SES and CIs for 

	:math:`h(E[f(Y(0), Y(1), X)], E[z_1(Y(1), X)], E[z_0(Y(0), X)])`.
	
	Parameters
	----------
	h : function
		real-valued function of fval, z0, z1, e.g.,
		``h = lambda fval, z0, z1 : fval / z0 + z1``.
	summands : np.array
		(2,n)-shaped array where summands[0].mean()
		is a lower bound on E[f(Y(0), Y(1), X)] and
		summands[1].mean() is an upper bound.
	z1summands : np.array
		(n,d)-length array where z1summands.mean(axis=0)
		estimates E[z_1(Y(1), X)].
	z0summands : np.array
		(n,d)-length array where z0summands.mean(axis=0) 
		estimates E[z_0(Y(0), X)].
	B : int
		Number of bootstrap replications.
	alpha : float
		Nominal level.
	"""
	d0 = z0summands.shape[1]
	# Stack everything in the appropriate format
	combined_summands = np.stack([
		np.concatenate([summands[[k]].T, z0summands, z1summands], axis=1) 
		for k in [0,1]
	], axis=0)
	return utilities.compute_est_bounds(
		summands=combined_summands,
		func=lambda x: h(
			fval=x[0], z0=x[1:(1+d0)], z1=x[(1+d0):]
		),
		B=B,
		clusters=clusters,
		alpha=alpha,
	)

class DeltaDualBounds(DualBounds):
	"""
	Computes generalized dual bounds via the delta method.

	The estimand is

	:math:`h(E[f(Y(0), Y(1), X)], E[z_1(Y(1), X)], E[z_0(Y(0), X)])`

	where h must be monotone increasing in its first argument. 
	
	Parameters
	----------
	h : function
		real-valued function of fval, z0, z1, e.g.,
		``h = lambda fval, z0, z1 : fval / z0 + z1``.
	z0 : function
		vector-valued function of y0, x.
	z1 : function
		vector-valued function of y1, x.
	f : function
		Function which defines the partially identified estimand.
		Must be a function of three arguments: y0, y1, x 
		(in that order). E.g.,
		``f = lambda y0, y1, x : y0 <= y1``
	outcome : np.array | pd.Series
		n-length array of outcome measurements (Y).
	treatment : np.array | pd.Series
		n-length array of binary treatment (W).
	covariates : np.array | pd.Series
		(n, p)-shaped array of covariates (X).
	propensities : np.array | pd.Series
		n-length array of propensity scores :math:`P(W=1 | X)`. 
		If ``None``, will be estimated from the data.
	clusters : np.array
		Optional n-length array of clusters, so ``clusters[i] = j``
		indicates that observation i is in cluster j.
	outcome_model : str | dist_reg.DistReg | list
		The model for estimating the law of :math:`Y | X, W`.
		Three options:

		- A str identifier, e.g., 'ridge', 'lasso', 'elasticnet', 'randomforest', 'knn'.
		- An object inheriting from ``dist_reg.DistReg``. 
		- A list of ``dist_reg.DistReg`` objects to automatically choose between.

		E.g., when ``outcome`` is continuous, the default is
		``outcome_model=dist_reg.CtsDistReg(model_type='ridge')``.
	propensity_model : str | sklearn classifier
		How to estimate the propensity scores if they are not provided.
		Two options:

		- A str identifier, e.g., 'ridge', 'lasso', 'elasticnet', 'randomforest', 'knn'.
		- An sklearn classifier, e.g., ``sklearn.linear_model.LogisticRegressionCV()``.
	model_selector : dist_reg.ModelSelector
		A ModelSelector object which can choose between several outcome models.
		The default performs within-fold nested cross-validation. Note: this
		argument is ignored unless ``outcome_model`` is a list.
	discrete : bool
		If True, treats the outcome as a discrete variable. 
		Defaults to ``None`` (inferred from the data).
	support : np.array
		Optional support of the outcome, if known and discrete.
		Defaults to ``None`` (inferred from the data).
	model_kwargs : dict
		Additional kwargs for the ``outcome_model``, e.g.,
		``feature_transform``. See 
		:class:`dualbounds.dist_reg.CtsDistReg` or 
		:class:`dualbounds.dist_reg.BinaryDistReg` for more kwargs.
	"""
	def __init__(
		self, 
		h: callable, 
		z1: callable, 
		z0: callable,
		*args,
		**kwargs,
	) -> None:
		self.h = h
		self.z1 = z1
		self.z0 = z0
		#self.h_grad = h_grad
		super().__init__(*args, **kwargs)

	def _plug_in_results(self, B: int=1000):
		ests, ses, cis = _delta_bootstrap_ses(
			h=self.h,
			summands=self.objvals,
			z1summands=self.z1summands,
			z0summands=self.z0summands,
			alpha=self.alpha,
			B=B,
			clusters=self.clusters,
		)
		return pd.DataFrame(
			np.stack(
				[ests, ses, cis], 
				axis=0
			),
			index=['Estimate', 'SE', 'Conf. Int.'],
			columns=['Lower', 'Upper']
		)

	def _compute_final_bounds(
		self, 
		aipw: bool=True,
		alpha: float=0.05,
		B: int=1000,
	):
		"""
		Computes final bounds based on (A)IPW summands,
		using the delta method or the bootstrap.

		Parameters
		----------
		aipw : bool
			If True, uses AIPW estimation. 
		alpha : float
			Nominal level.
		B : int
			Number of bootstrap replications. Default: 1000.
		"""
		# Compute summands for AIPW
		self._compute_ipw_summands()
		summands = self.aipw_summands if aipw else self.ipw_summands
		# Compute summands for z1 and z0
		self.z1_vals = np.stack(
			[self.z1(self.y[i], self.X[i]) for i in range(self.n)], axis=0
		).reshape(self.n, -1)
		d1 = self.z1_vals.shape[-1] # dimension of z1
		self.z0_vals = np.stack(
			[self.z0(self.y[i], self.X[i]) for i in range(self.n)], axis=0
		).reshape(self.n, -1)
		d0 = self.z0_vals.shape[-1] # dimension of z0

		### Step 1: use AIPW ideas to center/scale z1_values z0_vals
		if aipw:
			self.z1_mus = np.zeros((self.n, d1))
			self.z0_mus = np.zeros((self.n, d0))
			for i in range(self.n):
				nvals1 = self.y1_vals.shape[1]
				for j in range(nvals1):
					self.z1_mus[i] += self.z1(self.y1_vals[i, j], self.X[i]) * self.y1_probs[i, j]
				nvals0 = self.y0_vals.shape[1]
				for j in range(nvals0):
					self.z0_mus[i] += self.z0(self.y0_vals[i, j], self.X[i]) * self.y0_probs[i, j]
		else:
			self.z1_mus = 0
			self.z0_mus = 0

		Wr = self.W.reshape(-1, 1)
		pisr = self.pis.reshape(-1, 1)
		self.z1summands = (self.z1_vals - self.z1_mus) * Wr / pisr + self.z1_mus
		self.z0summands = (self.z0_vals - self.z0_mus) * (1 - Wr) / (1 - pisr) 
		self.z0summands += self.z0_mus

		## sample means and bootstrap
		# Bootstrap is valid because the analytical delta method is valid
		self.alpha = alpha
		self.estimates, self.ses, self.cis = _delta_bootstrap_ses(
			h=self.h,
			summands=summands,
			z1summands=self.z1summands,
			z0summands=self.z0summands,
			B=B,
			alpha=self.alpha,
			clusters=self.clusters,
		)
		# Return
		return dict(
			estimates=self.estimates,
			ses=self.ses,
			cis=self.cis,
		)



import time
import numpy as np
import pandas as pd
import scipy as sp
from scipy import stats
import unittest
import pytest
import os
import sys
try:
	from . import context
	from .context import dualbounds as db
# For profiling
except ImportError:
	import context
	from context import dualbounds as db

from dualbounds import varite, gen_data, delta


class TestDeltaDualBounds(context.DBTest):

	def test_deltadb_and_varite(self):
		""" This tests that DeltaDualBounds and VarITEDualBounds give the same answer. """
		data = gen_data.gen_regression_data(n=1000, p=5, dgp_seed=1, sample_seed=1)
		# Bootstrap variant
		deltadb  = delta.DeltaDualBounds(
			h=lambda fval, z1, z0: fval - (z1-z0)**2,
			z1=lambda y1, x: y1,
			z0=lambda y0, x: y0,
			f=lambda y0, y1, x: (y0-y1)**2,
			covariates=data['X'],
			outcome=data['y'],
			treatment=data['W'],
			propensities=data['pis'],
		)
		deltadb.fit().summary()
		# Analytical variant
		vdb = db.varite.VarITEDualBounds(
			covariates=data['X'],
			outcome=data['y'],
			treatment=data['W'],
			propensities=data['pis'],
		)
		vdb.fit(
			y0_dists=deltadb.y0_dists,
			y1_dists=deltadb.y1_dists,
			suppress_warning=True,
		).summary()
		for out, expected, name in zip(
			[deltadb.estimates, deltadb.cis],
			[vdb.estimates, vdb.cis],
			['estimates', 'cis'] 
		):
			np.testing.assert_array_almost_equal(
				out,
				expected,
				decimal=1,
				err_msg=f"VarITEDualBounds and DeltaDualBounds give different {name}"
			)

	def test_delta_from_pd(self):
		data = gen_data.gen_regression_data(n=50, p=5, eps_dist='bernoulli', r2=0)
		df = pd.DataFrame(data['X'])
		df['outcome'] = data['y']
		df['pis'] = data['pis']
		df['treatment'] = data['W']

		# Method 1
		fn_args = dict(
			f=lambda y0, y1, x: y1 - y0,
			h=lambda fval, z1, z0: fval,
			z0=lambda y0, x: y0,
			z1=lambda y1, x: y1, 
		)
		ddb1 = db.delta.DeltaDualBounds(
			**fn_args,
			covariates=pd.DataFrame(data['X']),
			outcome=pd.DataFrame(data['y']),
			treatment=pd.DataFrame(data['W']),
			propensities=pd.DataFrame(data['pis']),
		)
		# Method 2
		ddb2 = db.delta.DeltaDualBounds(
			**fn_args,
			covariates=data['X'],
			outcome=data['y'],
			treatment=data['W'],
			propensities=data['pis'],
		)

		# Test equality
		self.assertTrue(
			ddb1.X.shape == ddb2.X.shape,
			"initialization using pandas changes the shape of the covariates"
		)
		for expected, out, name in zip(
			[ddb2.y, ddb2.W, ddb2.pis],
			[ddb1.y, ddb1.W, ddb1.pis],
			['y', 'W', 'pis'],
		):
			np.testing.assert_array_almost_equal(
				expected,
				out,
				decimal=8,
				err_msg=f"DeltaDualBounds pandas init. changes {name} values."
			)

	def test_delta_clustered_ses(self):
		data = db.gen_data.gen_regression_data(
			n=200, p=1, r2=0.9, dgp_seed=1, sample_seed=1, eps_dist='gaussian',
			interactions=True,
		)
		f = lambda y0, y1, x: y1 * (y0 > 0)
		z0 = lambda y0, x: y0 > 0
		z1 = lambda y1, x: 0
		h = lambda fval, z0, z1: fval / z0
		# Stack data
		data = np.stack(
			[data['y'], data['W'], data['pis'], data['X'][:, 0]],
			axis=1
		)
		def ddb_se_function(data, clusters):
			ddb = db.delta.DeltaDualBounds(
				f=f, z0=z0, z1=z1, h=h,
				outcome=data[:, 0],
				treatment=data[:, 1],
				propensities=data[:, 2],
				covariates=data[:, 3].reshape(-1, 1),
				clusters=clusters,
			)
			ddb.fit().summary()
			return ddb.ses
		self.check_clustered_ses(func=ddb_se_function, data=data, msg_context='DeltaDualBounds')



if __name__ == "__main__":
	# Run all tests---useful if using cprofilev
	basename = os.path.basename(os.path.abspath(__file__))
	if sys.argv[0] == f'test/{basename}':
		time0 = time.time()
		context.run_all_tests([TestDeltaDualBounds()])
		elapsed = np.around(time.time() - time0, 2)
		print(f"Finished running all tests at time={elapsed}")

	# Else let unittest handle this for us
	else:
		unittest.main()


"""
Contains various interpolation functions used to 
go from discrete dual variables to continuous ones.
"""

import numpy as np
import scipy as sp
from . import utilities

def adaptive_interpolate(x: np.array, y: np.array, newx: np.array):
	"""
	Adaptively chooses between linear and nearest-neighbor
	interpolation.

	Parameters
	----------
	x : np.array
		n-length array of inputs. Must be sorted, although
		this is not explicitly enforced to save time.
	y : np.array
		n-length array of outputs
	newx : np.array
		m-length array of new inputs

	Returns
	-------
	newy : np.array
		m-length array of interpolated outputs
	"""
	if len(np.unique(y)) <= 2 and len(y) > 2:
		return nn_interpolate(x, y, newx)
	else:
		return linear_interpolate(x, y, newx)


def nn_interpolate(x: np.array, y: np.array, newx: np.array):
	"""
	Nearest-neighbor interpolation.

	Parameters
	----------
	x : np.array
		n-length array of inputs. Must be sorted, although
		this is not explicitly enforced to save time.
	y : np.array
		n-length array of outputs
	newx : np.array
		m-length array of new inputs

	Returns
	-------
	newy : np.array
		m-length array of interpolated outputs
	"""
	# Find nearest neighbors
	if not utilities.haslength(newx):
		newx = np.array([newx])
	n = len(x)
	rinds = np.minimum(np.searchsorted(x, newx, side='left'), n-1)
	linds = np.maximum(rinds-1, 0)
	inds = np.stack([linds, rinds], axis=1)
	dists = np.abs(x[inds] - newx.reshape(-1, 1))
	nbrs = inds[(np.arange(len(newx)), np.argmin(dists, axis=1))]
	# Return
	return y[nbrs]


def linear_interpolate(x: np.array, y: np.array, newx: np.array):
	"""
	Linear interpolation.

	Parameters
	----------
	x : np.array
		n-length array of inputs. Must be sorted, although
		this is not explicitly enforced to save time.
	y : np.array
		n-length array of outputs
	newx : np.array
		m-length array of new inputs

	Returns
	-------
	newy : np.array
		m-length array of interpolated outputs
	"""
	if not utilities.haslength(newx):
		newx = np.array([newx])
	# for now, check sorting (TODO DELETE)
	# if np.any(np.sort(x) != x):
	# 	raise ValueError("NOT SORTED")
	# interpolate points in the range of x
	haty = np.interp(newx, x, y)
	# Check if there are any points outside the boundaries
	lflags = newx < x[0]
	uflags = newx > x[-1]
	if not (np.any(lflags) or np.any(uflags)):
		return haty

	# Prevent div by 0 errors/warnings
	x, index = np.unique(x, return_index=True)
	y = y[index]

	# adjust for points < x.min()
	ldx = (y[1] - y[0]) / (x[1] - x[0])
	haty[lflags] = y[0] + (newx[lflags] - x[0]) * ldx
	# adjust for points > x.max()
	udx = (y[-1] - y[-2]) / (x[-1] - x[-2])
	haty[uflags] = y[-1] + (newx[uflags] - x[-1]) * udx
	return haty

def spline_interpolate(
    x, y, newx
):	
	if not utilities.haslength(newx):
		newx = np.array([newx])
	spline_rep = sp.interpolate.splrep(x=x, y=y)
	return sp.interpolate.splev(newx, spline_rep)


import copy
import os
import sys
import time
import numpy as np
import pandas as pd
from scipy import stats
from multiprocessing import Pool
from functools import partial
from itertools import product
from tqdm.auto import tqdm
from typing import Optional, Union

def elapsed(t0):
	return np.around(time.time() - t0, 2)

def vrange(n, verbose=False):
	if not verbose:
		return range(n)
	else:
		return tqdm(list(range(n)))

def haslength(x):
	try:
		len(x)
		return True
	except:
		return False

def floatable(x):
	try:
		float(x)
		return True
	except:
		return False

def itemize(x):
	try:
		return x.item()
	except AttributeError:
		return x

### For processing data
def _binarize_variable(x: np.array, var_name: str):
	"""
	Converts x to a binary r.v. and raises an error
	if it contains more than 2 values.
	"""
	# Check if nans
	if x.dtype in [int, np.int32, np.int64, np.float32, np.float64]:
		if np.any(np.isnan(x)):
			raise ValueError(f"{var_name} has missing values.")

	# Check if binary
	vals = set(list(np.unique(x)))
	if len(vals) > 2:
		raise ValueError(f"{var_name} is not binary and takes {len(vals)} values.")
	elif len(vals - set([0,1])) == 0:
		return x.astype(int)
	else:
		vals = sorted(list(vals))
		print(f"For {var_name}, replacing {vals[0]} with 0 and {vals[1]} with 1.")
		return (x == list(vals)[1]).astype(int)

def process_covariates(data: pd.DataFrame):
	"""
	Performs rudimentary pre-processing of covariates.

	Parameters
	----------
	data : pd.DataFrame
		Dataframe of covariates.

	Returns
	-------
	processed : pd.DataFrame
		DataFrame of preprocessed covariates.
	"""
	cts_covs = []
	discrete_covs = []
	for c in data.columns:
		if np.all(data[c].apply(floatable)):
			cts_covs.append(c)
		else:
			discrete_covs.append(c)

	# Get dummies and return
	return pd.get_dummies(
		data[discrete_covs + cts_covs],
		columns=discrete_covs,
		dummy_na=True,
		drop_first=True,
	).astype(float)

def preprocess_clusters(clusters):
	"""
	Parameters
	----------
	clusters : np.array
		n-length array of cluster indicators.

	Returns
	-------
	new_clusters : np.array
		n-length array of clusters which takes values
		from 0 to len(np.unique(clusters))-1.
	"""
	new_clusters = np.zeros(clusters.shape)
	for i, x in enumerate(np.sort(np.unique(clusters))):
		new_clusters[clusters == x] = i
	return new_clusters.astype(int)

def cluster_bootstrap_se(
	data: np.array,
	clusters: Optional[callable]=None,
	func: Optional[callable]=None,
	B: int=1000,
	verbose: bool=False,
):
	"""
	Computes clustered bootstrap on func(data).

	Parameters
	----------
	data : np.array
		array whose zeroth axis reflects the number of observations n.
		Typically a n-shaped or (n,d)-shaped array.
	func : callable
		A callable that maps an np.array to a scalar or array.
		Defaults to ``func=lambda x: x.mean()``.
	clusters : np.array
		n-shaped array where clusters[i] = j means that observation
		i is in the jth cluster.
	B : int
		Number of bootstrap draws to use. Ignored unless func or
		clusters are provided.
	verbose : bool
		If True, provides a progress bar.

	Returns
	-------
	se : float | np.array
		standard error of the estimator.
	bs_ests : np.array
		Array of bootstrapped estimators.
	"""
	# parse function
	n = len(data)
	if func is None:
		func = lambda x: x.mean()
	# Non-clustered case
	bs_ests = []
	if clusters is None:
		for b in vrange(B, verbose=verbose):
			inds = np.random.choice(n, n, replace=True)
			bs_ests.append(func(data[inds]))
	# Clustered case
	else:
		# Preprocess clusters
		clusters = preprocess_clusters(clusters).astype(int)
		n_clusters = len(np.unique(clusters))
		# data_clusters[j] = observations from cluster j
		data_clusters = [
			data[clusters==i] for i in range(n_clusters)
		]
		# Resample clusters uniformly at random
		for b in vrange(B, verbose=verbose):
			bs_clusters = np.random.choice(clusters, n_clusters, replace=True)
			new_data = np.concatenate(
				[data_clusters[i] for i in bs_clusters], axis=0
			)
			bs_ests.append(func(new_data))
	bs_ests = np.stack(bs_ests, axis=0)
	return bs_ests.std(axis=0), bs_ests

def compute_est_bounds(
	summands: np.array,
	clusters: Optional[callable]=None,
	func: Optional[callable]=None,
	B: int=1000,
	alpha: float=0.05,
):
	"""
	Helper to computes confidence intervals.

	Specifically, computes intervals for 

		``func(summands[k].mean(axis=-1))``

	Provides lower CI for k=0 and upper CI for k=1.

	Parameters
	----------
	summands : np.array
		(2, n)-shaped array or (2, n, d)-shaped array
	func : callable
		A callable that maps a 1D np.array to a scalar.
		Defaults to ``func=lambda x: x.mean()``.
	clusters : np.array
		n-shaped array where clusters[i] = j means that observation
		i is in the jth cluster.
	B : int
		Number of bootstrap draws to use. Ignored unless func or
		clusters are provided.
	alpha : float
		Nominal level.

	Returns
	-------
	ests : np.array
		2-shaped array of lower and upper estimators (sample mean).
	ses : np.array
		2-shaped array of standard errors.
	bounds : np.array
		2-shaped array of lower/upper confidence bounds.

	Notes
	-----
	This is meant primarily for internal use in the DualBounds class.
	"""
	scale = stats.norm.ppf(1-alpha/2)
	n = summands.shape[1]
	# In the simplest case, use analytical bounds
	if func is None and clusters is None:
		ests = summands.mean(axis=1)
		ses = summands.std(axis=1) / np.sqrt(n)
	# else, use the bootstrap/delta method
	else:
		if func is None:
			func = lambda x: x.mean()
		new_func = lambda data: itemize(func(data.mean(axis=0)))
		# Estimators
		ests = np.array([new_func(summands[k]) for k in [0,1]])
		ses = np.zeros(2)
		for k in [0,1]:
			ses[k] = itemize(cluster_bootstrap_se(
				data=summands[k],
				clusters=clusters,
				func=new_func,
				B=B,
			)[0])

	return ests, ses, np.array([
		ests[0] - scale * ses[0], ests[1] + scale * ses[1]
	])

def _sort_disc_dist(vals, probs):
	"""
	Parameters
	----------
	vals : (n, nvals)-shaped array.
	probs : (n, nvals)-shaped array. probs.sum(axis=1) == 1

	Returns
	-------
	vals_new : (n, nvals) array
		``vals`` sorted so that vals[i] is in increasing order.
	probs_new : (n, nvals) array
		``probs`` sorted in the corresponding order as vals.
	"""
	inds = np.argsort(vals, axis=1)
	new_vals = np.take_along_axis(vals, inds, axis=1)
	new_probs = np.take_along_axis(probs, inds, axis=1)
	return new_vals, new_probs

class ConstantDist:

	def __init__(self, loc=0, scale=1):
		self.loc = loc
		self.scale = scale

	def rvs(self, size):
		return self.loc + self.scale * np.ones(size)

class BatchedCategorical:
	"""
	Batched discrete (categorical) distribution.

	This class supports the small set of operations
	needed by the ``DualBounds`` class.

	Parameters
	----------
	vals : np.array
		(n, nvals)-shaped array of supports.
	probs : np.array
		(n, nvals)-shaped array of probabilities.
	"""
	def __init__(
		self, vals, probs
	):
		# sort
		self.n, self.nvals = vals.shape
		self.vals, self.probs = _sort_disc_dist(vals=vals, probs=probs)
		self.cumprobs = np.cumsum(self.probs, axis=1)
		# validate args
		if np.any(self.probs < 0):
			raise ValueError("probs must be nonnegative")
		if np.any(np.abs(self.probs.sum(axis=1)-1) > 1e-5):
			raise ValueError("probs.sum(axis=1) must equal 1")

	@classmethod
	def from_scipy(
		cls, scipy_bernoulli
	):
		"""
		Instantiates distribution from a 1D scipy.stats.bernoulli 
		object.

		Parameters
		----------
		scipy_dist : scipy.stats.bernoulli
		"""
		# Probabilities
		probs = scipy_bernoulli.mean()
		# Values 
		n = len(probs)
		vals = np.zeros((n, 2)); vals[:, 1] += 1
		# Return
		return cls(
			vals=vals, probs=np.stack([1-probs, probs], axis=1)
		)


	@classmethod
	def from_binary_probs(
		cls, probs
	):
		"""
		Instantiates distribution for a binary-valued variable.

		Parameters
		----------
		probs : np.array
			n-length array of P(Yi = 1)
		"""
		n = len(probs)
		probs_stacked = np.stack([1-probs, probs], axis=1)
		vals = np.zeros((n, 2))
		vals[:, 1] += 1
		return cls(vals=vals, probs=probs_stacked)
	
	def mean(self):
		"""
		Returns
		-------
		mu : np.array
			n-length mean of discrete distributions.
		"""
		return np.sum(self.vals * self.probs, axis=1)
	
	def ppf(self, q):
		"""
		Parameters
		----------
		q : np.array
			(m,n)-shaped array of desired quantiles, between 0 and 1.

		Returns
		-------
		quantile_values : np.array
			The desired quantile values of the categorical distributions.
		"""
		m = q.shape[0]
		# To be compatible with scipy, if q is (m, 1)
		# change it to (m,)
		if len(q.shape) == 2:
			if q.shape[1] == 1:
				q = q[:, 0]
		# Vectorize
		if len(q.shape) == 1:
			q = np.stack([q for _ in range(self.n)], axis=1)
		# use a for loop to save memory
		qvals = np.zeros((m, self.n))
		for i in range(self.n):
			flags = self.cumprobs[i].reshape(self.nvals, 1) >= q[:, i]
			qvals[:, i] = self.vals[i][np.argmax(flags, axis=0)]
		return qvals

	def rvs(self):
		"""
		Returns
		-------
		y : array
			n-length array of a single RV drawn from each categorical.
		"""
		u = np.random.uniform(size=(self.n, 1))
		inds = (u < self.cumprobs).argmax(axis=1)
		return self.vals[(np.arange(self.n), inds)]

def _adjust_support_size_unbatched(
	vals, probs, new_nvals, ymin, ymax
):
	"""
	Adjust categorical distribution to have a support of size new_nvals.
	
	Parameters
	----------
	vals : np.array
		nvals-length array of support. must be sorted.
	probs : np.array
		nvals-length array of probabilities
	new_nvals : int
		Desired size of support.
	ymin : float
		Minimum value for support
	ymax : float
		Maximum value for support
	"""
	nvals = len(vals)
	vals = vals.copy(); probs = probs.copy()
	if np.any(vals != np.sort(vals)):
		raise ValueError("vals must be sorted")
	# Case 0: no changes
	if nvals == new_nvals:
		return vals, probs
	# Case 1: just add padding
	elif nvals < new_nvals:
		pad = np.random.uniform(
			ymin, ymax, size=new_nvals - nvals
		)
		vals = np.concatenate([vals, pad], axis=0)
		probs = np.concatenate([probs, np.zeros(new_nvals - nvals)], axis=0)
		return vals, probs
	# Case 2: greedily merge support points
	else:
		probs = np.maximum(probs, 1e-8)
		probs /= probs.sum()
		while len(vals) > new_nvals:
			# Probability of equaling the left merge
			mixtures = probs[:-1] / (probs[:-1] + probs[1:])
			# candidate values for the post-merge support point
			candvals = vals[1:] * (1 - mixtures) + vals[:-1] * mixtures
			# E[|X - Y|] under best coupling
			costs = (candvals - vals[1:])**2 * probs[1:] 
			costs += (candvals - vals[:-1])**2 * probs[:-1]
			# pick the best variant
			istar = np.argmin(costs)
			vals[istar] = np.nan
			vals[istar + 1] = candvals[istar]
			probs[istar + 1] += probs[istar]
			probs[istar] = np.nan
			# drop na vals
			vals = vals[~np.isnan(vals)]
			probs = probs[~np.isnan(probs)]
		return vals, probs
	
def adjust_support_size(
	vals, probs, new_nvals, ymin, ymax
):
	"""
	Adjust categorical distribution to have support of size ``new_nvals``.
	
	Parameters
	----------
	vals : np.array
		(n, nvals)-length array of support. vals[i] must be sorted for each i.
	probs : np.array
		(n, nvals)-length array of probabilities of original distributions.
	new_nvals : int
		Desired size of support.
	ymin : float
		Minimum value for support.
	ymax : float
		Maximum value for support.

	Returns
	-------
	new_vals : np.array
		(n, new_nvals)-array of supports.
	new_probs : np.array
		(n, new_nvals)-array of probabilities.
	"""
	n, nvals = vals.shape
	new_vals = np.zeros((n, new_nvals))
	new_probs = np.zeros((n, new_nvals))
	for i in range(n):
		new_vals[i], new_probs[i] = _adjust_support_size_unbatched(
			vals[i], probs[i], new_nvals, ymin=ymin, ymax=ymax,
		)
	return new_vals, new_probs

def weighted_quantile(values, weights, quantiles, _old_style=True):
	"""
	Very close to numpy.percentile, but supports weights.

	Parameters
	----------
	values : np.array
		n-length array of data
	weights : np.array
		n-length array of sample weights
	quantiles : array-like
		d-length array of desired quantiles

	Returns
	-------
	results : np.array
		d-length array of weighted quantiles.
	"""
	if np.any(quantiles < 0) or np.any(quantiles > 1):
		raise ValueError("Quantiles must be in [0,1]")

	# Sort values
	sorter = np.argsort(values)
	values = values[sorter]
	weights = weights[sorter]
	# Compute quantiles
	cdf = np.cumsum(weights) - 0.5 * weights
	if _old_style:
		cdf -= cdf[0]
		cdf /= cdf[-1]
	else:
		cdf /= np.sum(weights)
	# Interp + return
	return np.interp(quantiles, cdf, values)

def parse_dist(
	dist, loc=0, scale=1, mu=None, sd=None, **kwargs
):
	## variant 1 based on mu/sd
	dist = dist.lower()
	if mu is not None and sd is not None and dist != 'bernoulli':

		# get SD correct
		temp = parse_dist(dist, loc=0, scale=sd, **kwargs)
		scale = sd / temp.std()
		temp = parse_dist(dist, loc=0, scale=sd*scale, **kwargs)
		shift = mu - temp.mean()
		return parse_dist(dist, loc=shift, scale=sd*scale, **kwargs)
	elif mu is not None and sd is not None and dist == 'bernoulli':
		# Note: Cannot enforce mean/sd for Bernoulli dist.
		# So we just use loc/scale.
		loc = copy.copy(mu)
		scale = copy.copy(sd)
		mu = None
		sd = None
	elif mu is not None or sd is not None:
		raise ValueError("Must provide both of mu and sd, not just one.")

	## variant 2 based on location/scale
	# sometimes return a regular dist object
	if not isinstance(dist, str):
		return dist

	# Parse
	if dist == 'constant':
		return ConstantDist(loc=loc, scale=scale)
	if dist == 'bernoulli':
		prob = np.exp(loc / scale)
		prob = prob / (1 + prob)
		return stats.bernoulli(p=prob)
	if dist in ['gaussian', 'normal']:
		return stats.norm(loc=loc, scale=scale)
	if dist == 'invchi2':
		df = kwargs.pop("df", 4)
		return stats.chi2(df=df, loc=loc, scale=scale/df)
	if dist == 't' or dist == 'tdist':
		df = kwargs.pop("df", 4)
		return stats.t(loc=loc, scale=scale, df=df)
	if dist == 'cauchy':
		return stats.cauchy(loc=loc, scale=scale)
	if dist == 'laplace':
		return stats.laplace(loc=loc, scale=scale)
	if dist in ['uniform', 'unif']:
		return stats.uniform(loc=loc-scale/2, scale=scale)
	if dist in ['expo', 'expon', 'exponential']:
		return stats.expon(loc=loc-scale, scale=scale)
	if dist == 'gamma':
		a = kwargs.pop("a", 5)
		return stats.gamma(loc=loc-a*scale, scale=scale, a=a)
	if dist == 'skewnorm':
		a = kwargs.pop("a", 5)
		return stats.skewnorm(loc=loc, scale=scale, a=a)
	if dist == 'powerlaw':
		a = kwargs.pop("a", 5)
		return stats.powerlaw(loc=loc, scale=scale, a=a)
	else:
		raise ValueError(f"Dist {dist} is unrecognized.")


"""
Methods for bounding Var(E[Y(1) - Y(0)]), the
variance of the individual treatment effect (ITE).
"""
import numpy as np
from scipy import stats
from .generic import DualBounds

def compute_analytical_varite_bound(
	n,
	y0_dists,
	y1_dists,
	reps=100,
):
	"""
	Semi-analytical bounds on :math:`Var(Y(0) - Y(1))`.

	Unlike dual bounds, this function is not
	robust to model misspecification.
	
	Parameters
	----------
	n : int
		Number of observations.
	y0_dists : stats.rv_continuous / stats.rv_discrete
		batched scipy distribution of shape (n,) where the ith
		distribution is the conditional law of Yi(0) | Xi
	y1_dists : stats.rv_continuous / stats.rv_discrete
		batched scipy distribution of shape (n,) where the ith
		distribution is the conditional law of Yi(1) | Xi
	reps : int
		Number of samples to take from each distribution.

	Returns
	-------
	lower : float
		Lower bound on Var(Y(1) - Y(0))
	upper : float
		Upper bound on Var(Y(1) - Y(0))
	"""
	# Sample coupled r.v.s
	U = np.random.uniform(size=(reps, n))
	y1 = y1_dists.ppf(U)
	y0l = y0_dists.ppf(U)
	y0u = y0_dists.ppf(1-U)
	# Evaluate bounds
	lower = np.std(y1-y0l)**2
	upper = np.std(y1 - y0u)**2
	return lower, upper

def varite_delta_method_se(
	sbetas, skappa1s, skappa0s
):
	# estimate
	hat_beta = sbetas.mean()
	hat_kappa1 = skappa1s.mean()
	hat_kappa0 = skappa0s.mean()
	ate = hat_kappa1 - hat_kappa0 # average treatment effect
	hattheta = hat_beta - ate**2
	# standard error
	hatSigma = np.cov(
		np.stack([sbetas, skappa1s, skappa0s], axis=0)
	) # 3 x 3 cov matrix
	grad = np.array(
		[1, - 2 * ate, 2 * ate]
	)
	# estimate
	se = np.sqrt(grad @ hatSigma @ grad / len(sbetas))
	return hattheta, se

class VarITEDualBounds(DualBounds):
	"""
	Computes dual bounds on :math:`Var(Y(1) - Y(0)).`

	The signature of this class is identical to 
	the ``generic.DualBounds`` class.  However, 
	the input ``f`` will be ignored.
	"""

	def __init__(self, *args, **kwargs):
		# Initialize with correct f function
		kwargs['f'] = lambda y0, y1, x: (y0-y1)**2
		super().__init__(*args, **kwargs)

	def results(self, minval: float=0, maxval: float=np.inf):
		# minval is always zero
		return super().results(minval=0, maxval=maxval)

	def _compute_final_bounds(self, aipw=True, alpha=0.05):
		"""
		Computes final bounds based in (A)IPW summands,
		using the delta method for Var(Y(1) - Y(0)).
		"""
		self._compute_ipw_summands()
		summands = self.aipw_summands if aipw else self.ipw_summands
		self._compute_cond_means()
		# Note: the notation follows Appendix A.2 of 
		# https://arxiv.org/pdf/2310.08115.pdf (version 1)
		ests = []
		ses = []
		bounds = []
		scale = stats.norm.ppf(1-alpha/2)
		for lower in [1, 0]:
			# part. identifiable component
			sbetas = summands[1-lower]
			# kappa1 = E[Y(1)], kappa0 = E[Y(0)] are ident components
			skappa1s = self.W * (self.y - self.mu1) / self.pis + self.mu1
			skappa0s = (1-self.W) * (self.y - self.mu0) / (1-self.pis) + self.mu0
			# estimate
			hattheta, se = varite_delta_method_se(
				sbetas=sbetas, skappa1s=skappa1s, skappa0s=skappa0s
			)
			ests.append(hattheta)
			ses.append(se)
			if lower:
				bounds.append(hattheta - scale * se)
			else:
				bounds.append(hattheta + scale * se)

		self.estimates = np.array(ests)
		self.ses = np.array(ses)
		self.cis = np.array(bounds)
		return dict(
			estimates=self.estimates,
			ses=self.ses,
			cis=self.cis
		)


import time
import numpy as np
import pandas as pd
import scipy as sp
from scipy import stats
import unittest
import pytest
import os
import sys
import sklearn.ensemble
from sklearn import linear_model as lm
try:
	from . import context
	from .context import dualbounds as db
# For profiling
except ImportError:
	import context
	from context import dualbounds as db

from dualbounds import generic, iv, utilities, gen_data
from dualbounds.utilities import parse_dist, BatchedCategorical


def _balke_pearl_bounds(wprobs, ydists):
	"""
	wprobs : np.array
		wprobs[i, z] = P(W_i(z) = 1)
	ydists : list
		ydists[z][w] is a n-shaped scipy distribution
		representing the distributions of Y_i(w) | W_i(z) = w, X_i.
		These must be Bernoulli distributions.
	"""
	# Convert to probabilities
	yprobs = [[], []]
	for z in [0,1]:
		for w in [0,1]:
			yprobs_zw = ydists[z][w].mean()
			if np.any(yprobs_zw < 0) or np.any(yprobs_zw > 1):
				raise ValueError("y must be binary to apply balke_pearl bounds")
			yprobs[z].append(yprobs_zw)

	def compute_pi(y, w, z):
		"""
		Computes vector of probabilities of P(Y=y, W=w | Z = z)
		This is pi in https://arxiv.org/pdf/2301.12106.
		"""

		## 1. vector of probabilities P(W = w) 
		wp = (1-w) * (1 - wprobs[:, z]) + w * wprobs[:, z]

		## 2. joint probabilities
		if y == 1:
			return wp * yprobs[z][w]
		else:
			return wp * (1 - yprobs[z][w])

	# Compute balke pearl bounds
	# This is ugly, but I'm not sure of a better solution than hard-coding
	### (a) minimum
	thetal1 = compute_pi(1, 1, 1) + compute_pi(0, 0, 0) - 1
	thetal2 = compute_pi(1, 1, 0) + compute_pi(0, 0, 1) - 1
	thetal3 = - compute_pi(y=0, w=1, z=1) - compute_pi(y=1, w=0, z=1)
	thetal4 = - compute_pi(y=0, w=1, z=0) - compute_pi(y=1, w=0, z=0)
	thetal5 = compute_pi(1, 1, 0) - compute_pi(1, 1, 1) - compute_pi(1, 0, 1) - compute_pi(0, 1, 0) - compute_pi(1, 0, 0)
	thetal6 = compute_pi(1, 1, 1) - compute_pi(1, 1, 0) - compute_pi(1, 0, 0) - compute_pi(0, 1, 1) - compute_pi(1, 0, 1)
	thetal7 = compute_pi(0, 0, 1) - compute_pi(0, 1, 1) - compute_pi(1, 0, 1) - compute_pi(0, 1, 0) - compute_pi(0, 0, 0)
	thetal8 = compute_pi(0, 0, 0) - compute_pi(0, 1, 0) - compute_pi(1, 0, 0) - compute_pi(0, 1, 1) - compute_pi(0, 0, 1)
	gammal = np.stack(
		[thetal1, thetal2, thetal3, thetal4, thetal5, thetal6, thetal7, thetal8],
		axis=0
	).max(axis=0)

	### (b) maximum
	thetau1 = 1 - compute_pi(0, 1, 1) - compute_pi(1, 0, 0)
	thetau2 = 1 - compute_pi(0, 1, 0) - compute_pi(1, 0, 1)
	thetau3 = compute_pi(1, 1, 1) + compute_pi(0, 0, 1)
	thetau4 = compute_pi(1, 1, 0) + compute_pi(0, 0, 0)
	thetau5 = - compute_pi(0, 1, 0) + compute_pi(0, 1, 1) + compute_pi(0, 0, 1) + compute_pi(1, 1, 0) + compute_pi(0, 0, 0)
	thetau6 = - compute_pi(0, 1, 1) + compute_pi(1, 1, 1) + compute_pi(0, 0, 1) + compute_pi(0, 1, 0) + compute_pi(0, 0, 0)
	thetau7 = - compute_pi(1, 0, 1) + compute_pi(1, 1, 1) + compute_pi(0, 0, 1) + compute_pi(1, 1, 0) + compute_pi(1, 0, 0)
	thetau8 = - compute_pi(1, 0, 0) + compute_pi(1, 1, 0) + compute_pi(0, 0, 0) + compute_pi(1, 1, 1) + compute_pi(1, 0, 1)
	gammau = np.stack(
		[thetau1, thetau2, thetau3, thetau4, thetau5, thetau6, thetau7, thetau8],
		axis=0
	).min(axis=0)
	return gammal, gammau

class TestDualIVBounds(unittest.TestCase):
	
	def test_balke_pearl_reduction(self):
		"""
		Test that dualIV bounds reduce to Balke-Pearl for binary y.
		""" 
		n = 100
		data = db.gen_data.gen_iv_data(
			n=n, p=2, tau=0.5, eps_dist='bernoulli', dgp_seed=1, sample_seed=1,
		)
		
		# Balke-Pearl bounds
		bp_lower, bp_upper = _balke_pearl_bounds(
			wprobs=data['wprobs'], ydists=data['ydists']
		)
		bp_bounds = np.stack([bp_lower, bp_upper], axis=0)
		
		# Plug-in bounds
		dbiv = iv.DualIVBounds(
			f=lambda w0, w1, y0, y1, x: y1 - y0,
			outcome=data['y'],
			exposure=data['W'],
			instrument=data['Z'],
			propensities=data['pis'],
			covariates=None,
			discrete=True,
			support=set([0,1]),
			suppress_iv_warning=True,
		)
		ydists_cat = [[np.nan, np.nan], [np.nan, np.nan]]
		for z in [0,1]:
			for w in [0,1]:
				ydists_cat[z][w] = BatchedCategorical.from_scipy(data['ydists'][z][w])

		dbiv.compute_dual_variables(
			wprobs=data['wprobs'],
			ydists=ydists_cat,
			verbose=True,
		)
		np.testing.assert_array_almost_equal(
			dbiv.objvals,
			bp_bounds,
			decimal=5,
			err_msg="For binary y, DualIVBounds objvals do not match Balke-Pearl bounds"
		)
		# Test relationship with CATE
		self.assertTrue(
			np.all(dbiv.objvals[0] <= data['cates']),
			"lower objvals are sometimes larger than the CATE"
		)
		self.assertTrue(
			np.all(dbiv.objvals[1] >= data['cates']),
			"upper objvals are sometimes smaller than the CATE"
		)

		# Test that the AIPW process works. First, sample new data
		reps = 2000
		newZ = np.random.binomial(1, data['pis'], size=(reps, n))
		newW = np.random.binomial(1, data['wprobs'][(np.arange(n), newZ)])
		newY = np.zeros((reps, n)) - 10
		for z in [0,1]:
			for w in [0,1]:
				ysamp = data['ydists'][z][w].rvs(size=(reps, n))
				newY[(newW == w) & (newZ == z)] = ysamp[(newW == w) & (newZ == z)]

		# Now resample 
		aipw_estimators = np.zeros((reps, 2))
		ipw_estimators = np.zeros((reps, 2))
		for r in range(reps):
			dbiv._compute_realized_dual_variables(
				y=newY[r], W=newW[r], Z=newZ[r],
			)
			dbiv.Z = newZ[r]
			dbiv.W = newW[r]
			dbiv.y = newY[r]
			dbiv._compute_ipw_summands()
			dbiv._compute_final_bounds(aipw=False)
			aipw_estimators[r] = dbiv.estimates
			dbiv._compute_final_bounds(aipw=True)
			ipw_estimators[r] = dbiv.estimates
		# Test equality
		for estimators, name in zip([ipw_estimators, aipw_estimators], ['ipw', 'aipw']):
			np.testing.assert_array_almost_equal(
				estimators.mean(axis=0),
				bp_bounds.mean(axis=1),
				decimal=2,
				err_msg=f"DualIVBounds {name} estimators do not match Balke-Pearl bounds"
			)

	def test_late_reduction(self):
		"""
		Check that DualIVBounds under monotonicity can estimate the LATE.
		"""
		n = 20
		data = db.gen_data.gen_iv_data(
			n=n, 
			p=2,
			dgp_seed=1,
			sample_seed=1,
		)
		# DBIV with monotonicity
		dbiv = iv.DualIVBounds(
			f=lambda w0, w1, y0, y1, x: (y1 - y0) * w1 * (1-w0),
			outcome=data['y'],
			exposure=data['W'],
			instrument=data['Z'],
			propensities=data['pis'],
			covariates=None,
			support_restriction=lambda w0, w1, y0, y1, x: w0 <= w1,
			suppress_iv_warning=True,
		)
		dbiv.compute_dual_variables(
			wprobs=data['wprobs'],
			ydists=data['ydists'],
			verbose=True,
		)
		# For simplicity, just check numerator (later will use DeltaDualIVBounds)
		late_numerators = 0
		for z in [0,1]:
			for w in [0,1]:
				wp = w * data['wprobs'][:, z] + (1 - w) * (1 - data['wprobs'][:, z])
				late_numerators += (2*z - 1) * data['ydists'][z][w].mean() * wp
		# Check that we get the right results
		for lower, name in zip([1, 0], ['lower', 'upper']):
			np.testing.assert_array_almost_equal(
				dbiv.objvals[1-lower],
				late_numerators,
				decimal=5,
				err_msg=f"{name} DualIV objective values do not match LATE numerator under monotonicity"
			)

	# def test_consistency(self):
	# 	"""
	# 	Check that the final bounds are similar to the oracle bounds.
	# 	"""
	# 	pass



if __name__ == "__main__":
	# Run all tests---useful if using cprofilev
	basename = os.path.basename(os.path.abspath(__file__))
	if sys.argv[0] == f'test/{basename}':
		time0 = time.time()
		context.run_all_tests([TestDualIVBounds()])
		elapsed = np.around(time.time() - time0, 2)
		print(f"Finished running all tests at time={elapsed}")

	# Else let unittest handle this for us
	else:
		unittest.main()



"""
Methods for bounding Var(E[Y(1) - Y(0) | X]).
"""
import warnings
import numpy as np
from scipy import stats
from . import generic, utilities
from .generic import infer_discrete
import pandas as pd
import copy
from typing import Optional

def _moments2varcate(
	hxy1, hxy0, hx, y1, y0, shx2,
):
	"""
	Computes bound on the variance of the
	CATE from a set of marginal moments
	on E[h(X) Y(1)], E[h(X) Y(0)], E[h(X)],
	etc.
	"""
	return 2 * (
		hxy1 - hxy0 - hx * y1 + hx * y0 
	) - shx2 + hx**2

def varcate_delta_method_se(
	shxy1, shxy0, shx, sy1, sy0, shx2,
):
	"""
	Estimates and provides SE for 
	2 * Cov(h(X), Y(1) - Y(0)) - var(h(X)).

	Parameters
	----------
	shxy1 : np.array
		n-length array of AIPW summands for E[Y(1) h(X)].
	shxy0 : np.aray
		n-length array of AIPW summands for E[Y(0) h(X)].
	shx : np.array
		n-length array of AIPW summands for E[h(X)].
	sy1 : np.array
		n-length array of AIPW summands for E[Y(1)].
	sy0 : np.array
		n-length array of AIPW summands for E[Y(1)].
	shx2 : np.array
		n-length array of AIPW summands for E[h(X)^2].

	Returns
	-------
	estimate : float
		Plug-in estimate
	se : float
		Standard error
	"""
	# concatenate and estimate
	summands = np.stack(
		[shxy1, shxy0, shx, sy1, sy0, shx2], axis=1
	)
	mus = summands.mean(axis=0)
	hattheta = _moments2varcate(
		*tuple(list(mus))
	)
	hatSigma = np.cov(summands.T)
	grad = np.array([
		2,
		-2, 
		2 * (mus[4] - mus[3] + mus[2]),
		- 2 * mus[2],
		2 * mus[2],
		-1,
	])
	# estimate
	se = np.sqrt(grad @ hatSigma @ grad / len(shx))
	return hattheta, se

class VarCATEDualBounds(generic.DualBounds):
	"""
	Computes lower bounds on :math:`Var(E[Y(1) - Y(0) | X]).`

	This class has the same signature as ``generic.DualBounds``
	except it only provides lower bounds and the input ``f``
	will be ignored.
	"""

	def __init__(self, *args, **kwargs):
		# Initialize with no f function
		kwargs['f'] = None
		# change default eps_dist for cts to Gaussian
		# (eps_dist has zero impact on the final results here 
		# and Gaussian is the most computationally efficient)
		kwargs['eps_dist'] = 'gaussian'
		# Initialize
		super().__init__(*args, **kwargs)

	def _ensure_feasibility(self):
		raise NotImplementedError()

	def _discretize(self):
		raise NotImplementedError()

	def _solve_single_instance(self):
		raise NotImplementedError()

	def _compute_realized_dual_variables(self):
		raise NotImplementedError()

	def compute_dual_variables(self, *args, **kwargs):
		""" 
		In this case, the optimal dual variables are simply 
		the estimated CATE, so this function does nothing.
		"""
		pass

	def _compute_ipw_summands(self):
		pass

	def diagnostics(self, plot=True, aipw=True):
		return None

	def results(self, minval: float=0, maxval: float=np.inf):
		# minval is always zero
		return super().results(minval=0, maxval=maxval)

	def _plug_in_results(self, B: int=1000):
		# Estimates
		pests = np.array([self.cates.std()**2, np.nan])
		# SEs via bootstrap
		inds = np.random.choice(self.n, size=(self.n, B), replace=True)
		bootstrap_varcates = self.cates[inds].std(axis=0)**2
		pses = np.array([bootstrap_varcates.std(), np.nan])
		pcis = pests - pses * stats.norm.ppf(1-self.alpha) 
		return pd.DataFrame(
			np.stack(
				[pests, pses, pcis], 
				axis=0
			),
			index=['Estimate', 'SE', 'Conf. Int.'],
			columns=['Lower', 'Upper']
		)

	def _compute_final_bounds(self, aipw=True, alpha=0.05):
		self._compute_cond_means()
		self.cates = self.mu1 - self.mu0 
		#### We have to use the 6-d delta method
		# The notation uses h(X) = hat E[Y(1) - Y(0) | X]
		# 1. AIPW terms for h(X) * Y(1)
		self.shxy1 = self.W * self.cates * (self.y - self.mu1)
		self.shxy1 = self.shxy1 / self.pis + self.cates * self.mu1
		# 2. AIPW terms for h(X) * Y(0)
		self.shxy0 = (1 - self.W) * self.cates * (self.y - self.mu0)
		self.shxy0 = self.shxy0 / (1 - self.pis) + self.cates * self.mu0
		# 3. AIPW terms for h(X)
		self.shx = self.cates 
		# 4. AIPW terms for Y(1)
		self.sy1 = self.W * (self.y - self.mu1) / self.pis + self.mu1
		# 5. AIPW terms for Y(0)
		self.sy0 = (1 - self.W ) * (self.y - self.mu0)
		self.sy0 = self.sy0 / (1 - self.pis) + self.mu0
		# 6. AIPW terms for h(X)^2
		self.shx2 = self.cates**2
		estimate, se = varcate_delta_method_se(
			shxy1=self.shxy1, 
			shxy0=self.shxy0, 
			shx=self.shx, 
			sy1=self.sy1, 
			sy0=self.sy0, 
			shx2=self.shx2,
		)
		scale = stats.norm.ppf(1-alpha)
		lower_ci = estimate - scale * se
		self.estimates = np.array([estimate, np.nan])
		self.ses = np.array([se, np.nan])
		self.cis = np.array([lower_ci, np.nan])

def varcate_cluster_bootstrap(
	varcate_objects: list[VarCATEDualBounds],
	aipw: bool=True,
	alpha: float=0.05,
	B: int=1000,
	verbose: bool=False,
):
	"""
	Combines evidence across multiple VarCATEDualBounds classes
	using a (clustered) bootstrap.

	Parameters
	----------
	varcate_objects : list
		A list of fit VarCATEDualBounds objects.
	alpha : float
		Nominal level, between 0 and 1.
	B : int
		Number of bootstrap replications.
	verbose : bool
		If True, prints a progress report.
	"""
	K = len(varcate_objects)
	# Initial estimates and ses
	estimates = np.stack([x.estimates for x in varcate_objects], axis=-1)[0] # K-length array
	ses = np.stack([x.ses for x in varcate_objects], axis=-1)[0] # K-length array
	# Create data: n x 2 x K
	data = np.stack(
		[np.stack([vdb.cates, vdb.sy1 - vdb.sy0], axis=1) for vdb in varcate_objects],
		axis=2
	)
	# Function which maps data to maximum VarCATE estimate
	def compute_estimates(data):
		K = data.shape[-1]
		ests = np.zeros(K)
		for k in range(K):
			cov_truecate = np.cov(data[:, 0, k], data[:, 1, k])[0,1]
			var_estcate = np.std(data[:, 0, k])**2
			ests[k] = 2 * cov_truecate - var_estcate
		return ests
	# Cluster bootstrap
	_, bs_estimators = utilities.cluster_bootstrap_se(
		data=data,
		clusters=varcate_objects[0].clusters,
		func=compute_estimates,
		B=B,
		verbose=verbose,
	) # B x K
	# Centered BS estimators and quantile
	centered = (bs_estimators - estimates) / ses
	hatq = np.quantile(centered.max(axis=-1), 1-alpha)
	# Take maximums
	optimal_index = np.argmax(estimates - hatq * ses)
	combined_estimates = np.maximum(np.array([estimates[optimal_index], np.nan]), 0)
	final_ses = np.array([ses[optimal_index], np.nan])
	cis = np.maximum(np.array([(estimates - hatq * ses)[optimal_index], np.nan]), 0)
	# Return
	return pd.DataFrame(
		np.stack([combined_estimates, final_ses, cis], axis=0),
		index=['Estimate', 'SE', 'Conf. Int.'],
		columns=['Lower', 'Upper'],
	)

DEFAULT_SHRINKAGE_VALUES = np.array([0.001, 0.01, 0.1, 0.3, 0.5, 0.7, 0.9, 1, 1.5, 2])

class CalibratedVarCATEDualBounds(VarCATEDualBounds):
	"""
	Improved lower bounds on :math:`Var(E[Y(1) - Y(0) | X])`.

	This has the same signature as ``varcate.VarCATEDualBounds`` 
	except it takes one additional argument. 

	Parameters
	----------
	shrinkages : np.array
		Array of shrinkage values s, where the CATEs are replaced with
		(1-s) * CATE + s * ATE. 

	Notes
	-----
	This class is identical to ``varcate.VarCATEDualBounds``
	except it selects a shrinkage parameter which shrinks
	estimated CATEs towards the ATE. Also, if ``outcome_model`` is a list,
	it uses the multiplier bootstrap to perform model selection. 

	This class should not be used for observational studies.
	"""
	def __init__(self, *args, shrinkages=DEFAULT_SHRINKAGE_VALUES, **kwargs):
		self.shrinkages = shrinkages
		# initialize outcome models
		self.outcome_models = kwargs.pop("outcome_model", ['ridge'])
		if not isinstance(self.outcome_models, list):
			self.outcome_models = [self.outcome_models]
		# create vdbs
		self.vdbs = []
		for outcome_model in self.outcome_models:
			self.vdbs.append(
				VarCATEDualBounds(*args, **kwargs, outcome_model=outcome_model)
			)

	def fit(self, B: int=1000, **kwargs):
		# Fit all models
		verbose = kwargs.get("verbose", True)
		counter = 0
		for outcome_model, vdb in zip(self.outcome_models, self.vdbs):
			if verbose and len(self.vdbs) > 1:
				print(f"Fitting model={outcome_model}, num. {counter} out of {len(self.vdbs)}.")
				counter += 1
			vdb.fit(**kwargs)
		
		# Run multiplier bootstrap at various shrinkages
		self.shrunk_vdb_objects = []
		for vdb in self.vdbs:	
			hatate = np.mean(
				vdb.mu1 - vdb.mu0
				+ vdb.W*(vdb.y - vdb.mu1)/vdb.pis
				- (1-vdb.W) * (vdb.y - vdb.mu0)/ (1-vdb.pis)
			)
			mu0_homogeneous = (vdb.mu0 + vdb.mu1) / 2 - hatate/2
			mu1_homogeneous = (vdb.mu0 + vdb.mu1) / 2 + hatate/2
			aipw = vdb.sy1 - vdb.sy0
			for s in self.shrinkages:
				new_mu0 = vdb.mu0 * s + (1 - s) * mu0_homogeneous
				new_mu1 = vdb.mu1 * s + (1 - s) * mu1_homogeneous
				new_vdb = VarCATEDualBounds(
					outcome=vdb.y,
					treatment=vdb.W,
					propensities=vdb.pis,
				).fit(
					y0_dists=stats.norm(loc=new_mu0),
					y1_dists=stats.norm(loc=new_mu1),
					suppress_warning=True,
				)
				self.shrunk_vdb_objects.append(new_vdb)

		# Cluster bootstrap
		if verbose:
			print("Fitting cluster bootstrap to aggregate results.")
		results = varcate_cluster_bootstrap(
			self.shrunk_vdb_objects, B=B, verbose=verbose, alpha=self.vdbs[0].alpha
		)
		self.estimates = results.loc['Estimate'].values
		self.ses = results.loc['SE'].values
		self.cis = results.loc['Conf. Int.'].values
		return self


	def summary(self, minval: Optional[float]=0, maxval: Optional[float]=np.inf):
		print("___________________Inference_____________________")
		print(self.results(minval=minval, maxval=maxval))
		print()
		print("_________________Outcome models__________________")
		# Loop through and print model diagnostics
		for outcome_model, vdb in zip(self.outcome_models, self.vdbs):
			print(f"Outcome model: {outcome_model}")
			print(vdb.eval_outcome_model())




import os
import sys
import numpy as np
import unittest 

# Add path to allow import of code
file_directory = os.path.dirname(os.path.abspath(__file__))
parent_directory = os.path.split(file_directory)[0]
sys.path.insert(0, os.path.abspath(parent_directory))

# Import the actual stuff
import dualbounds

# for profiling
import inspect

def _run_se_computation_test(
	dim,
	f,
	se_function,
	arg_names,
	n=500,
	reps=10000,
	testname='',
	param_shift=0,
	decimal=2,
):
	"""
	Generic function to test whether a delta method computation
	is correct.

	dim : int
		Number of parameters
	f : function
		f(*params) = target of inference 
	se_function : function
		function which computes SES given vectors of data
	arg_names : list
		List of length dim of input names for se_function 
	n : int
		Number of data-points per dataset
	reps : int
		Number of datasets
	shift : float
		params ~ N(param_shift, 1).
	decimal : positive float
		Number of decimals to ensure accuracy
	"""
	np.random.seed(123)
	# Create parameters
	mu = np.random.randn(dim) + param_shift
	sigmas = np.random.uniform(0.1, 0.9, size=dim)
	theta = f(*mu)
	# sample data
	data = np.random.randn(dim, reps, n) #+ mu.reshape(-1, 1, 1)
	data *= sigmas.reshape(-1, 1, 1)
	data += mu.reshape(-1, 1, 1)
	# Loop through and compute estimators
	ests = np.zeros(reps); ses = np.zeros(reps)
	for r in range(reps):
		kwargs = {
			arg_names[d]:data[d, r] for d in range(dim)
		}
		est, se = se_function(**kwargs)
		ests[r] = est
		ses[r] = se
	# test approximate unbiasedness
	np.testing.assert_array_almost_equal(
		ests.mean(),
		theta,
		decimal=decimal,
		err_msg=f"{testname} plug-in estimator is not unbiased"
	)
	# test SE estimation
	np.testing.assert_array_almost_equal(
		ests.std(),
		ses.mean(),
		decimal=decimal,
		err_msg=f"{testname} SE estimator is not consistent"
	)

class DBTest(unittest.TestCase):

	def check_clustered_ses(
		self,
		func: callable,
		data: np.array,
		msg_context: str='',
		rel_tol=0.2,
	):
		"""
		Checks that SEs do not change much if we concatenate the
		same dataset to itself 5 times and use clustered standard errors.

		Parameters
		----------
		func : callable
			function of two arguments: ``data`` and ``clusters``
			which returns an array of standard errors.
		data : np.array
			(n,d)-shaped array of data.
		msg_context : str
			context to print upon an error
		rel_tol : float
			Ensures new_ses / ses <= 1 + rel_tol and ses / new_ses <= 1 + rel_tol
		"""
		ses = func(data, clusters=None)
		k = 4
		n = len(data)
		# Create new dataset, where we just repeat observations
		new_data = np.concatenate([data for _ in range(k)], axis=0)
		clusters = np.concatenate([np.arange(n) for _ in range(k)], axis=0)
		inds = np.argsort(clusters)
		clusters = clusters[inds]
		new_data = new_data[inds]
		# Create new ses
		new_ses = func(new_data, clusters=clusters)
		for obj1, obj2 in zip([ses, new_ses], [new_ses, ses]):
			self.assertTrue(
				np.all(obj1 / obj2 <= 1 + rel_tol),
				msg=f"For {msg_context}, ses={ses} change if have size-{k} clusters, yielding new_ses={new_ses}"
			)


def run_all_tests(test_classes):
	"""
	Usage: 
	context.run_all_tests(
		[TestClass(), TestClass2()]
	)
	This is useful for making pytest play nice with cprofilev.
	"""
	def is_test(method):
		return str(method).split(".")[1][0:4] == 'test'
	for c in test_classes:
		attrs = [getattr(c, name) for name in c.__dir__()]
		test_methods = [
			x for x in attrs if inspect.ismethod(x) and is_test(x)
		]
		for method in test_methods:
			method()